## @module PopulationModel
# Returns the current population (number of living farmbits)
# @param levels: Levels applicable for model

from typing import List, Optional, Dict, Any
import logging
import datetime

from models.SequenceModel import SequenceModel
from ogd.core.processors.LakelandExtractor import  LakelandExtractor

# birth/death events
# ["FARMBITDEATH", "NEWFARMBIT"]
# [18, 31]

## @class PopulationModel
# Returns the current population (number of living farmbits)
# @param levels: Levels applicable for model
class PopulationModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class PopulationModel
        Returns the current population (number of living farmbits)
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, rows: List[Dict[str, Any]]) -> int:
        try:
            pop = self._eval_assert(rows)
        except AssertionError:
            return -1
        return pop

    def _eval_assert(self, events: List[Dict[str, Any]]) -> int:
        # start_time = datetime.datetime.fromisoformat(events[0]["client_time"])
        skip = True
        assert events
        population = 0
        for event in events:
            # If the event is a newfarmbit, add 1 to population.
            # If the event is a farmbitdeath, subtract 1 from the population.
            event_custom = event["event_custom"]
            event_str = LakelandExtractor._ENUM_TO_STR['EVENT CATEGORIES'][event_custom]
            if event_str not in ["emote"]:
                # cur_time = datetime.datetime.fromisoformat(event["client_time"])
                debug_str = ''
                if event_str == "buy" and event["event_data_complex"]["success"]:
                    debug_str = LakelandExtractor._ENUM_TO_STR['BUYS'][event["event_data_complex"]["buy"]].upper()
                if event_str == 'startgame':
                    debug_str = "START" if not event["event_data_complex"].get("continue") else "CONTINUE"
                    debug_str += f' Language: {event["event_data_complex"].get("language")}'
                if event_str in ['availablefood', 'sadfarmbit']:
                    true_pop = event["event_data_complex"].get("farmbit")
                    debug_str = f'true pop = {true_pop}'
                    assert population == true_pop, f"Expected {population} farmbits, but true_pop is {true_pop}"

                event_str = f'{event_str} {debug_str}'
                # logging.debug(f"{population:<3} {event_str:<20} {cur_time-start_time}")
            if skip:
                if event_custom not in [1, 31]: # newgames start with STARTGAME; continues start with NEWFARMBIT
                    continue
                else:
                    skip = False
            if event_custom == 31: # newfarmbit
                population += 1
            elif event_custom == 18: # death
                population = population - 1
                assert population >= 0, "Population became negative"
            elif event_custom in [40,23]: # reset / endgame
                skip = True
                population = 0
        return population


    def __repr__(self):
        return f"PopulationModel(levels={self._levels}, input_type={self._input_type})"



# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.games.ICECUBE.DBExport import scenes_list
## @class PerLevelFeature
class PerSceneFeature(PerCountFeature):
    """PerLevelFeature
    Abstract base class for per-level game features.
    Works like a normal Feature, but checks if the given event has right "level"
    before attempting to extract from event.

    Args:
        Feature (_type_): PerLevelFeature is a subclass of Feature

    Returns:
        _type_: _description_
    """

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _validateEventCountIndex(self, event:Event):
        return event.EventData['scene_name'] == scenes_list[self.CountIndex]


## import standard libraries
import json
from pathlib import Path
from typing import Any, Callable, Dict, Final, List, Optional
## import local files
from . import features
from ogd.games import PENGUINS
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.PENGUINS.detectors import *
from ogd.games.PENGUINS.features import *
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema
# from ogd.games.PENGUINS.DBExport import scene_map

## @class WaveExtractor
#  Extractor subclass for extracting features from Waves game data.

EXPORT_PATH : Final[str] = "games/PENGUINS/DBExport.json"

class PenguinsLoader(GeneratorLoader):


    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the WaveLoader class.
    def __init__(self, player_id:str, session_id: str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]=None):
        """Constructor for the WaveLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        self._region_map : List[Dict[str, Any]] = []

        # Load Penguins jobs export and map job names to integer values
        _dbexport_path = Path(PENGUINS.__file__) if Path(PENGUINS.__file__).is_dir() else Path(PENGUINS.__file__).parent
        with open(_dbexport_path / "DBExport.json", "r") as file:
            export = json.load(file)
            self._region_map = export.get("regions", [])

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @staticmethod
    def _getFeaturesModule():
        return features
    
    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        if extractor_params._count_index == None:
            match feature_type:
                case "ActivityCompleted":
                    ret_val = ActivityCompleted.ActivityCompleted(params=extractor_params)
                case "ActivityDuration":
                    ret_val = ActivityDuration.ActivityDuration(params=extractor_params)       
                case "BuiltNestCount":
                    ret_val = BuiltNestCount.BuiltNestCount(params=extractor_params)
                case "BuiltWrongNestCount":
                    ret_val = BuiltWrongNestCount.BuiltWrongNestCount(params=extractor_params)
                case "EatFishCount":
                    ret_val = EatFishCount.EatFishCount(params=extractor_params)
                case "EggLostCount":
                    ret_val = EggLostCount.EggLostCount(params=extractor_params)
                case "EggRecoverTime":
                    ret_val = EggRecoverTime.EggRecoverTime(params=extractor_params)
                case "GazeCount":
                    ret_val = GazeCount.GazeCount(params=extractor_params)
                case "GazeDuration":
                    ret_val = GazeDuration.GazeDuration(params=extractor_params)
                case "LogVersion":
                    ret_val = LogVersion.LogVersion(params=extractor_params)
                case "MirrorWaddleDuration":
                    ret_val = MirrorWaddleDuration.MirrorWaddleDuration(params=extractor_params)
                case "PenguinInteractCount":
                    ret_val = PenguinInteractCount.PenguinInteractCount(params=extractor_params)
                # case "PlayerInactiveAvgDuration":
                #     ret_val = PlayerInactiveAvgDuration.PlayerInactiveAvgDuration(params=extractor_params)
                case "RegionsEncountered":
                    ret_val = RegionsEncountered.RegionsEncountered(params=extractor_params)
                case "RingChimesCount":
                    ret_val = RingChimesCount.RingChimesCount(params=extractor_params)
                case "RockBashCount":
                    ret_val = RockBashCount.RockBashCount(params=extractor_params) 
                case "RockPickupCount":
                    ret_val = RockPickupCount.RockPickupCount(params=extractor_params)
                case "RockMultiplePickupCount":
                        ret_val = RockMultiplePickupCount.RockMultiplePickupCount(params=extractor_params)      
                case "SessionDuration":
                    ret_val = SessionDuration.SessionDuration(params=extractor_params, session_id=self._session_id)
                case "SkuaBashCount":
                        ret_val = SkuaBashCount.SkuaBashCount(params=extractor_params)
                case "SkuaPeckCount":
                        ret_val = SkuaPeckCount.SkuaPeckCount(params=extractor_params)
                case "SnowBallDuration":
                    ret_val = SnowBallDuration.SnowBallDuration(params=extractor_params)
                case "WaddleCount":
                    ret_val = WaddleCount.WaddleCount(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid aggregate feature for Penguins.")
        # Per-count features
        # level attempt features
        else:
            match feature_type:
                case "RegionDuration":
                            ret_val = RegionDuration.RegionDuration(params=extractor_params, region_map=self._region_map)
                case "RegionEnterCount":
                        ret_val = RegionEnterCount.RegionEnterCount(params=extractor_params, region_map=self._region_map)
                case "WaddlePerRegion":
                    ret_val = WaddlePerRegion.WaddlePerRegion(params=extractor_params, region_map=self._region_map)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid per-count feature for Penguins.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        ret_val : Detector

        match detector_type:
            case "RegionEnter":
                ret_val = RegionEnter.RegionEnter(params=extractor_params, trigger_callback=trigger_callback, region_map=self._region_map)
            case "RegionExit":
                ret_val = RegionExit.RegionExit(params=extractor_params, trigger_callback=trigger_callback, region_map=self._region_map)
            case _:
                raise NotImplementedError(f"'{detector_type}' is not a valid detector for Penguins.")
        return ret_val



        # Load Aqualab scenes export and map scene names to integer values
        # with open(EXPORT_PATH, "r") as file:
        #     export = json.load(file)

        #     task_num = 1
        #     for i, scene in enumerate(export["scenes"], start=1):
        #         self._scene_map[scene["id"]] = i
        #         self._diff_map[i] = scene["difficulties"]
        #         for task in scene["tasks"]:
        #             task_by_scene = scene["id"] + "_" + task["id"]
        #             self._task_map[task_by_scene] = task_num
        #             task_num += 1

        # Update level count
        # self._game_schema._max_level = len(self._scene_map) - 1


## import standard libraries
import logging
import typing
## import local libraries
from models.FeatureModel import FeatureModel

## @class SingleFeatureModel
#  Implementation for models which simply mirror a single feature to the dashboard.
#  Dead simple "model," but convenient way to integrate individual features into the dashboard.
class SingleFeatureModel(FeatureModel):
    def __init__(self, column_name: str, levels: typing.List[int] = []):
        super().__init__(levels)
        self._col_name = column_name

    ## Abstract declaration of a function to perform calculation of a model results from a row.
    #
    #  @param rows : A row of data for a session, which should be a mapping of column names to 
    #  @return     : A result for the given row of data
    def _eval(self, row: typing.Dict):
        return row[self._col_name]


# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class HeadsetOnCount(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._headset_count : int = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["headset_on"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        self._headset_count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._headset_count]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "2"


from typing import List, Optional, Dict, Any
from models.SequenceModel import SequenceModel

## @class MoneyAccumulationModel
# Returns a list with (1) the current amount of money and (2) the total number of homes,
# farms, and dairy farms
# Rationale: These two values give a simple-to-interpret snapshot of how much money someone is
# hanging onto compared to how well-developed their town is. If money is high, but the town is
# small, it could represent indecision in the player.
# @param levels: Levels applicable for model
class MoneyAccumulationModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class MoneyAccumulationModel
        Returns a list with (1) the current amount of money and (2) the total number of homes,
        farms, and dairy farms
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> List[int]:
        assert events
        homes = 0
        farms = 0
        dairy = 0
        money = 0
        money_since_gamestate = 0
        for event in reversed(events):
            # tally money for selling stuff since most recent gamestate event
            if event["event_custom"] == 37:
                money_since_gamestate += event["event_data_complex"]["worth"]
            # get tallies from most recent gamestate event
            if event["event_custom"] == 0:
                money = event["event_data_complex"]["money"]
                tile_string = event["event_data_complex"]["tiles"]
                _tile = self._read_stringified_array(tile_string)
                tile = self._array_to_mat(4, _tile)
                for i in range(0, len(tile)):
                    if tile[i][3] == 8:
                        homes += 1
                    elif tile[i][3] == 9:
                        farms += 1
                    elif tile[i][3] == 10:
                        dairy += 1
                break
        return [money + money_since_gamestate, homes + farms + dairy]

    # reformat raw variable functions
    def _read_stringified_array(self, arr: str) -> List[int]:
        if not arr:
            return []
        return [int(x) for x in arr.split(',')]


    def _array_to_mat(self, num_columns: int, arr: List[int]) -> List[List[int]]:
        assert len(arr) % num_columns == 0
        return [arr[i:i + num_columns] for i in range(0, len(arr), num_columns)]


def __repr__(self):
        return f"MoneyAccumulationModel(levels={self._levels}, input_type={self._input_type})"


# import libraries
from datetime import datetime
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobLocationChanges(PerJobFeature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        self._job_map = job_map
        super().__init__(params=params, job_map=job_map)
        self._by_task       : Dict[str, int] = {}
        self._total_count   : int = 0
        self._current_count : int = 0
        self._last_time     : Optional[datetime] = None
        self._last_type     : str = "NO_CHANGE"

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["scene_change", "room_change", "complete_task"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "scene_change":
            if self._validateSceneChange(event=event):
                self._current_count += 1
                self._total_count += 1
                self._last_type = event.EventName
                self._last_time = event.Timestamp
        elif event.EventName == "room_change":
            if self._validateRoomChange(event=event):
                self._current_count += 1
                self._total_count += 1
                self._last_type = event.EventName
                self._last_time = event.Timestamp
        elif event.EventName == "complete_task":
            task_id = event.EventData.get("task_id", "UNKNOWN_TASK")
            self._by_task[task_id] = self._current_count
            self._current_count = 0

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._total_count, self._by_task]

    # *** Optionally override public functions. ***

    def Subfeatures(self) -> List[str]:
        return ["ByTask"]

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"

    # *** Private Methods ***

    def _validateSceneChange(self, event:Event):
        if self._last_type == "scene_change":
            return True # if last thing was a scene change, this must be a new change.
        elif self._last_type == "room_change":
            if self._last_time is not None and abs((self._last_time - event.Timestamp).total_seconds()) < 0.01:
                return False # if last thing was a room change, and occurred at effectively the same time, then this is a doubling-up of the two, don't count it.
            else:
                return True # if we haven't encountered a previous time, or the two were separated by some meaningful time, this is a new change.
        else:
            return True # if we didn't encounter a valid previous type of change, then this is a new change

    def _validateRoomChange(self, event:Event):
        if self._last_type == "room_change":
            return True # if last thing was a room change, this must be a new change.
        elif self._last_type == "scene_change":
            if self._last_time is not None and abs((self._last_time - event.Timestamp).total_seconds()) < 0.01:
                return False # if last thing was a scene change, and occurred at effectively the same time, then this is a doubling-up of the two, don't count it.
            else:
                return True # if we haven't encountered a previous time, or the two were separated by some meaningful time, this is a new change.
        else:
            return True # if we didn't encounter a valid previous type of change, then this is a new change



## import standard libraries
import abc
import json
import logging
from datetime import datetime
from typing import Any, Dict, List, Type, Optional
# import locals
from ogd.core.models.FeatureData import FeatureData
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.Event import Event
from ogd.core.utils.utils import ExportRow, Logger

## @class Processor
class Processor(abc.ABC):

    # *** ABSTRACTS ***

    ## Abstract declaration of a function to get the calculated value of the feature, given data seen so far.
    @abc.abstractmethod
    def _processEvent(self, event:Event) -> None:
        pass

    @abc.abstractmethod
    def _getLines(self) -> List[ExportRow]:
        pass

    @abc.abstractmethod
    def _clearLines(self) -> None:
        pass

    # *** BUILT-INS ***

    def __init__(self, game_schema: GameSchema):
        self._game_schema : GameSchema = game_schema

    def __str__(self):
        return f"Processor object for {self._game_schema.GameName} data"

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def ProcessEvent(self, event:Event) -> None:
        # TODO: add error handling code, if applicable.
        self._processEvent(event=event)
    
    @property
    def Lines(self) -> List[ExportRow]:
        ret_val : List[ExportRow] = []

        # TODO: Should do list of events instead, and put it on outerface to do this part.
        # When retrieving lines, ensure all lines have values in appropriate format for output to file/web response/whatever.
        _dumps_default = lambda x : x.isoformat() if isinstance(x, datetime) else str(x)
        ret_val = [[json.dumps(_col, default=_dumps_default) for _col in _line] for _line in self._getLines()]

        return ret_val

    def ClearLines(self):
        self._clearLines()

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import standard libraries
import math
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.detectors.DetectorEvent import DetectorEvent
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode

class RegionEnter(Detector):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters, trigger_callback:Callable[[Event], None], region_map:List[Dict[str, Any]]):
        super().__init__(params=params, trigger_callback=trigger_callback)
        self._region_map = region_map
        self._last_timestamp = datetime.now()
        self._old_region : str = "NoRegion"
        self._new_region : str = "NoRegion"

    # *** Implement abstract functions ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["player_waddle"]

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        old_position : Dict[str, float]
        new_position : Dict[str, float]

        int_version = int(event.LogVersion) # Penguins just uses an int for the log_version
        match int_version:
            case int_version if int_version <= 9:
                old_position = {
                    'x': event.GameState.get('old_posX', -math.inf),
                    'y': event.GameState.get('old_posY', -math.inf),
                    'z': event.GameState.get('old_posZ', -math.inf)
                }
                new_position = {
                    'x': event.GameState.get('posX', -math.inf),
                    'y': event.GameState.get('posY', -math.inf),
                    'z': event.GameState.get('posZ', -math.inf)
                }
            # New format, up to latest version as of last change to file:
            case int_version if int_version <= 11:
                _pos_old = event.GameState.get("pos_old", [-math.inf, -math.inf, -math.inf])
                _pos_new = event.GameState.get("pos_new", [-math.inf, -math.inf, -math.inf])
                old_position = {
                    'x': _pos_old[0],
                    'y': _pos_old[1],
                    'z': _pos_old[2],
                }
                new_position = {
                    'x': _pos_new[0],
                    'y': _pos_new[1],
                    'z': _pos_new[2],
                }
            # Default to current format as of last change to file:
            case _:
                _pos_old = event.GameState.get("pos_old", [-math.inf, -math.inf, -math.inf])
                _pos_new = event.GameState.get("pos_new", [-math.inf, -math.inf, -math.inf])
                old_position = {
                    'x': _pos_old[0],
                    'y': _pos_old[1],
                    'z': _pos_old[2],
                }
                new_position = {
                    'x': _pos_new[0],
                    'y': _pos_new[1],
                    'z': _pos_new[2],
                }

        # Need to set to "NoRegion" as the default.
        self._old_region = "NoRegion"
        self._new_region = "NoRegion"

        for region in self._region_map:
            if (old_position['x'] > region['minX'] and 
                old_position['x'] < region['maxX'] and
                old_position['y'] > region['minY'] and
                old_position['y'] < region['maxY'] and
                old_position['z'] > region['minZ'] and
                old_position['z'] < region['maxZ']):
                self._old_region = region['name']
            if (new_position['x'] > region['minX'] and 
                new_position['x'] < region['maxX'] and
                new_position['y'] > region['minY'] and
                new_position['y'] < region['maxY'] and
                new_position['z'] > region['minZ'] and
                new_position['z'] < region['maxZ']):
                self._new_region = region['name']
        self._last_timestamp = event.Timestamp
    
    def _trigger_condition(self) -> bool:
        """_summary_
        """
        if (self._old_region != self._new_region) and (self._new_region != "NoRegion"):
            return True
        else:
            return False

    def _trigger_event(self) -> Optional[Event]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        ret_val : Event = self.GenerateEvent(app_id="PENGUINS", timestamp=self._last_timestamp,
                                             event_name="region_enter", event_data={"region":self._new_region})
        return ret_val

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import json
from typing import Any, Final, List, Optional
import json
from datetime  import timedelta, datetime
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature





class SessionPlayTime(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    IDLE_TIME_THRESHOLD : Final[timedelta] = timedelta(seconds=60)

    def __init__(self, params:GeneratorParameters, threshold: int):
        super().__init__(params=params)
        self._first_event_timestamp : Optional[datetime] = None
        self._last_event_timestamp : Optional[datetime] = None

        self._idle_time : Optional[timedelta] = None
        self._time : timedelta = timedelta(0)
        self._count : int = 0
        self._last_timestamp : Optional[datetime] = None
        self._threshold : timedelta = timedelta(seconds=threshold)
        self._total_time : Optional[timedelta] = None

        # >>> create/initialize any variables to track feature extractor state <<<
        #
        # e.g. To track whether extractor found a click event yet:
        # self._found_click : bool = False
    
    @staticmethod
    def defaultThreshold():
        return SessionPlayTime.IDLE_TIME_THRESHOLD

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["all_events"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
        if not self._first_event_timestamp:
            self._first_event_timestamp = event.Timestamp

        if not self._last_timestamp:
            self._last_timestamp = event.Timestamp
            print("set self._last!")
            return
        if self._last_timestamp is not None:
            time_since_last = event.Timestamp - self._last_timestamp
            if time_since_last > SessionPlayTime.IDLE_TIME_THRESHOLD:
                self._time += time_since_last
                self._count += 1
        else:
            raise ValueError("In IdleState, last timestamp is None!")
        self._last_timestamp = event.Timestamp

        if(not self._first_event_timestamp):
            self._first_event_timestamp = event.Timestamp
        
        self._last_event_timestamp = event.Timestamp


        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        # print(f'_first_timestamp is:{self._first_event_timestamp}, _last_timestamp is: {self._last_timestamp}')
        
        # if(not self._first_event_timestamp):
        #     self._total_time = 0
        #     play_time : int = 0
        # else:
        #     self._total_time = self._last_timestamp-self._first_event_timestamp
        #     play_time : timedelta = self._total_time - self._time

        #total idle time, idle count, total session time, (total session time - idle time)
        # return [self._time, self._count, self._total_time, play_time]
        total_time : timedelta = self._last_event_timestamp - self._first_event_timestamp

        return [total_time, 0, 0, 0]


    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Count", "Total Time", "Total Play Time"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



## import standard libraries
import bisect
import logging
import typing
from datetime import datetime
from typing import Any, Dict, List, Optional
## import local files
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.legacy.LegacyFeature import LegacyFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

# temp comment

## @class CrystalExtractor
#  Extractor subclass for extracting features from Crystal game data.
class CrystalExtractor(LegacyFeature):
    ## Constructor for the CrystalExtractor class.
    #  Initializes some custom private data (not present in base class) for use
    #  when calculating some features.
    #  Sets the sessionID feature.
    #  Further, initializes all Q&A features to -1, representing unanswered questions.
    #
    #  @param session_id The id number for the session whose data is being processed
    #                    by this extractor instance.
    #  @param game_table A data structure containing information on how the db
    #                    table assiciated with this game is structured. 
    #  @param game_schema A dictionary that defines how the game data itself is
    #                     structured.
    def __init__(self, params:GeneratorParameters, game_schema:GameSchema, session_id:str):
        super().__init__(params=params, game_schema=game_schema, session_id=session_id)
        # Define custom private data.
        self._start_times: typing.Dict       = {}
        self._end_times:   typing.Dict       = {}
        self._totalMoleculeDragDuration      = {}
        self._active_begin = None
        self._features.setValByName(feature_name="sessionID", new_value=session_id)
        # we specifically want to set the default value for questionAnswered to -1, for unanswered.
        for ans in self._features.getValByName(feature_name="questionAnswered").keys():
            self._features.setValByIndex(feature_name="questionAnswered", index=ans, new_value=None)
        for q in self._features.getValByName(feature_name="questionCorrect"):
            self._features.setValByIndex(feature_name="questionCorrect", index=q, new_value=None)

    ## Function to perform extraction of features from a row.
    #
    #  @param row_with_complex_parsed A row of game data from the db, with the
    #                                 "complex data" already parsed from JSON.
    #  @param game_table  A data structure containing information on how the db
    #                     table assiciated with this game is structured.
    def _updateFromEvent(self, event:Event):
        # put some data in local vars, for readability later.
        level = event.GameState['level']
        event_client_time = event.Timestamp
        # Check for invalid row.
        if self.ExtractionMode == ExtractionMode.SESSION and event.SessionID != self._session_id:
            Logger.Log(f"Got a row with incorrect session id! Expected {self._session_id}, got {event.SessionID}!", logging.ERROR)
        # If row is valid, process it.
        else:
            # If we haven't set persistent id, set now.
            if self._features.getValByName(feature_name="persistentSessionID") == 0:
                self._features.setValByName(feature_name="persistentSessionID", new_value=event.UserData['persistent_session_id'])
            # Ensure we have private data initialized for this level.
            if not level in self._levels:
                bisect.insort(self._levels, level)
                self._features.initLevel(level)
                self._totalMoleculeDragDuration[level] = 0
                # self.start_times[level] = None
                # self.end_times[level] = None
            # 1) record that an event of any kind occurred, for the level & session
            self._features.incValByIndex(feature_name="eventCount", index=level)
            self._features.incAggregateVal(feature_name="sessionEventCount")
            # 2) figure out what type of event we had. If CUSTOM, we'll use the event_custom sub-item.
            event_type = event.EventName.split('.')[0]
            if event_type == "CUSTOM":
                event_type = event.EventData['event_custom']
            # 3) handle cases for each type of event
            match event_type:
                case "BEGIN":
                    self._extractFromBegin(level, event_client_time)
                case "COMPLETE":
                    self._extractFromComplete(level, event_client_time, event.EventData)
                case "MOLECULE_RELEASE":
                    self._extractFromMoleculeRelease(level, event.EventData)
                case "MOLECULE_ROTATE":
                    self._extractFromMoleculeRotate(level, event.EventData)
                case "BACK_TO_MENU":
                    self._extractFromMenuBtn(level, event_client_time)
                case "CLEAR_BTN_PRESS":
                    self._extractFromClearBtnPress(level)
                case "MUSEUM_CLOSE" :
                    self._extractFromMuseumClose(event.EventData)
                case "QUESTION_ANSWER":
                    self._extractFromQuestionAnswer(event.EventData)
                case _:
                    raise Exception(f"Found an unrecognized event type: {event_type}")
                                               
    ## Function to perform calculation of aggregate features from existing
    #  per-level/per-custom-count features.
    def _calculateAggregateFeatures(self):
        # Calculate per-level averages and percentages, since we can't calculate
        # them until we know how many total events occur.
        for level in self._levels:
            count = self._features.getValByIndex(feature_name="moleculeMoveCount", index=level)
            avg = self._totalMoleculeDragDuration[level] / count if (count is not None and count > 0) else count
            self._features.setValByIndex(feature_name="avgMoleculeDragDurationInSecs", index=level, new_value=avg)

    ## Private function to extract features from a "BEGIN" event.
    #  The features affected are:
    #  - start_times (used to calculate durationInSecs and sessionDurationInSecs)
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    def _extractFromBegin(self, level, event_client_time:datetime):
        self._features.incValByIndex(feature_name="beginCount", index=level, increment=1)
        if self._active_begin == None:
            self._start_times[level] = event_client_time
        elif self._active_begin == level:
            pass # in this case, just keep going.
        else:
            self._end_times[self._active_begin] = event_client_time
            time_taken = self._calcLevelTime(self._active_begin)
            self._features.incValByIndex(feature_name="durationInSecs", index=self._active_begin, increment=time_taken)
            self._features.incAggregateVal(feature_name="sessionDurationInSecs", increment=time_taken)
            self._start_times[level] = event_client_time
        # in any case, current level now has active begin event.
        self._active_begin = level

    ## Private function to extract features from a "COMPLETE" event.
    #  The features affected are:
    #  - end_times (used to calculate durationInSecs and sessionDurationInSecs)
    #  - durationInSecs
    #  - sessionDurationInSecs
    #  - completesCount
    #  - finalScore
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    #  @param event_data        Parsed JSON data from the row being processed.
    def _extractFromComplete(self, level, event_client_time:datetime, event_data:Dict[str,Any]):
        self._features.incValByIndex(feature_name="completesCount", index=level, increment=1)
        if self._active_begin == None:
            sess_id = self._features.getValByName(feature_name="sessionID")
            Logger.Log(f"Got a 'Complete' event when there was no active 'Begin' event! Level {level}, Sess ID: {sess_id}", logging.ERROR)
        else:
            self._end_times[level] = event_client_time
            time_taken = self._calcLevelTime(level)
            self._features.incValByIndex(feature_name="durationInSecs", index=level, increment=time_taken)
            self._features.incAggregateVal(feature_name="sessionDurationInSecs", increment=time_taken)
            self._active_begin = None
            score = event_data["stability"]["pack"] + event_data["stability"]["charge"]
            max_score = max(score, self._features.getValByIndex(feature_name="finalScore", index=level))
            self._features.setValByIndex(feature_name="finalScore", index=level, new_value=max_score)

    ## Private function to extract features from a "BACK_TO_MENU" event.
    #  The features affected are:
    #  - end_times (used to calculate totalLevelTime and avgLevelTime)
    #  - durationInSecs
    #  - sessionDurationInSecs
    #  - menuBtnCount
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    def _extractFromMenuBtn(self, level:int, event_client_time:datetime):
        self._features.incValByIndex(feature_name="menuBtnCount", index=level)
        if self._active_begin == None:
            sess_id = self._features.getValByName(feature_name="sessionID")
            Logger.Log(f"Got a 'Back to Menu' event when there was no active 'Begin' event! Sess ID: {sess_id}", logging.ERROR)
        else:
            self._end_times[level] = event_client_time
            time_taken = self._calcLevelTime(level)
            self._features.incValByIndex(feature_name="durationInSecs", index=level, increment=time_taken)
            self._features.incAggregateVal(feature_name="sessionDurationInSecs", increment=time_taken)
            self._active_begin = None

    ## Private function to extract features from a "MOLECULE_RELEASE" event.
    #  The features affected are:
    #  - totalMoleculeDragDuration (used to calculate avgMoleculeDragDurationInSecs)
    #  - moleculeMoveCount
    #  - sessionMoleculeMoveCount
    #
    #  @param level      The level being played when event occurred.
    #  @param event_data Parsed JSON data from the row being processed.
    def _extractFromMoleculeRelease(self, level:int, event_data:Dict[str,Any]):
        if not level in self._totalMoleculeDragDuration.keys():
            self._totalMoleculeDragDuration[level] = 0
        self._totalMoleculeDragDuration[level] += event_data["time"]
        self._features.incValByIndex(feature_name="moleculeMoveCount", index=level)
        self._features.incAggregateVal(feature_name="sessionMoleculeMoveCount")

    ## Private function to extract features from a "MOLECULE_ROTATE" event.
    #  The features affected are:
    #  - stampRotateCount
    #  - sessionStampRotateCount
    #  - singleRotateCount
    #  - sessionSingleRotateCount
    #
    #  @param level      The level being played when event occurred.
    #  @param event_data Parsed JSON data from the row being processed.
    def _extractFromMoleculeRotate(self, level:int, event_data:Dict[str,Any]):
        if event_data["isStamp"]:
            self._features.incValByIndex(feature_name="stampRotateCount", index=level)
            self._features.incAggregateVal(feature_name="sessionStampRotateCount")
        else:
            self._features.incValByIndex(feature_name="singleRotateCount", index=level)
            self._features.incAggregateVal(feature_name="sessionSingleRotateCount")

    ## Private function to extract features from a "CLEAR_BTN_PRESS" event.
    #  The features affected are:
    #  - clearBtnPresses
    #  - sessionClearBtnPresses
    #
    #  @param level The level being played when clear button was pressed.
    def _extractFromClearBtnPress(self, level:int):
        self._features.incValByIndex(feature_name="clearBtnPresses", index=level)
        self._features.incAggregateVal(feature_name="sessionClearBtnPresses")

    ## Private function to extract features from a "MUSEUM_CLOSE" event.
    #  The features affected are:
    #  - sessionMuseumDurationInSecs
    #
    #  @param event_data Parsed JSON data from the row being processed.
    def _extractFromMuseumClose(self, event_data:Dict[str,Any]):
        self._features.incAggregateVal(feature_name="sessionMuseumDurationInSecs", increment=event_data["timeOpen"])

    ## Private function to extract features from a "QUESTION_ANSWER" event.
    #  The features affected are:
    #  - questionAnswered
    #  - questionCorrect
    #
    #  @param event_data_parsed Parsed JSON data from the row being processed.
    def _extractFromQuestionAnswer(self, event_data:Dict[str,Any]):
        q_num = event_data["question"]
        self._features.setValByIndex(feature_name="questionAnswered", index=q_num, new_value=event_data["answered"])
        correctness = 1 if event_data["answered"] == event_data["answer"] else 0
        self._features.setValByIndex(feature_name="questionCorrect", index=q_num, new_value= correctness)

    ## Private function to calculate the time spent on a level for the given level
    #  and play index (each level may have been played multiple times).
    #
    #  @param lvl      The level whose play time should be calculated
    #  @return         The play time if level is valid, -1 if level was not
    #                  completed, 0 if level was never attempted.
    def _calcLevelTime(self, lvl:int) -> int:
        # use 0 if not played, -1 if not completed
        if lvl in self._levels:
            if lvl in self._start_times and lvl in self._end_times:
                return (self._end_times[lvl] - self._start_times[lvl]).total_seconds()
            else:
                return -1
        else:
            return 0


## @module FeatSeqPercentModel
# Feature to output the percentile that a session is at along a progression of gameplay timestamp features. Depends
# heavily on the logic used to produce the quantiles file. The following exemplifies current implementation:
#
# Example: A game has 10 checkpoints. Joey has reached 3/10 checkpoints. His total playtime so far is 45 seconds.
# The model will return the percentile of reaching checkpoint 4 at 45 seconds. If historically 70% of students take
# longer than 45 seconds to reach checkpoint 4, the model will return 30%.
#
# Example: A game has 10 checkpoints. Joey has reached 10/10 checkpoints. His total playtime so far is 190 seconds.
# He reached checkpoint 10 at 185 seconds. The model will return the percentile of reaching checkpoint 10 at 185
# seconds, and will continue returning that same value for this session.
#
# @param feature_sequence: sequence of timedelta features that are to be in strictly ascending order
# @param levels: levels applicable to this model
# @param time_feat: "sessDuration" or "sess_time_active", depending if the model should be used for active time or
# overall time. Make sure that the quantiles file uses the same time feature.
# @param quantile_json_path: Path to a quantiles JSON file constructed by the _FeatureQuantiles private class


from typing import List, Optional, Dict, Any
import pandas as pd
from bisect import bisect_left
import numpy as np
import json
from datetime import timedelta
from models.FeatureModel import FeatureModel
from ogd.core.utils.Logger import Logger

_POP_ACHS = "exist group town city".split()
_FARM_ACHS = "farmer farmers farmtown megafarm".split()
_MONEY_ACHS = "paycheck thousandair stability riches".split()
_BLOOM_ACHS = "bloom bigbloom hugebloom massivebloom".split()
_REQ_TUTORIALS = "buy_food build_a_farm timewarp \
successful_harvest sell_food buy_fertilizer buy_livestock \
livestock poop rain".split()  # skip build_a_house - it comes at 0.0


def _get_sess_active_time_to_achievement_list(achs: List[str]) -> List[str]:
    return [f'sess_time_active_to_{a}_achievement' for a in achs]


def _get_sess_active_time_to_tutorial_list(tuts: List[str]) -> List[str]:
    return [f'sess_time_active_to_{t}_tutorial' for t in tuts]


def _get_quantiles(df: pd.DataFrame, feats: List[str], filter_debug:bool=True, filter_continue:bool=True) -> Dict[str, List[float]]:
    filter_strings = []
    if filter_debug:
        filter_strings += ['(debug==0)']
    if filter_continue:
        filter_strings += ['(c==0)']
    if filter_strings:
        df = df.rename({"continue": "c"}, axis=1).query(' & '.join(filter_strings)).rename({"c": "continue"}, axis=1)
    df = df[feats].replace(0.0, pd.NA)
    df = df.quantile(np.arange(0, 1, .01))
    quantiles = df.to_dict('list')
    return quantiles

## @class _FeatureQuantiles
# A private class used to create singleton access to quantile data used by the FeatSeqPercentModel class.
class _FeatureQuantiles(object):

    def __init__(self, arg, filter_continue=True):
        if type(arg) is str:
            json_path = arg
            with open(json_path) as f:
                self._quantiles = json.load(f)
            return

        df = arg
        cols = df.select_dtypes(include="number").columns
        self._quantiles = _get_quantiles(df, cols, filter_continue=filter_continue)

    @classmethod
    def fromDF(cls, df: pd.DataFrame, filter_continue=True) -> 'FeatureQuantiles':
        return cls(df, filter_continue=filter_continue)

    @classmethod
    def fromCSV(cls, csv_path: str, filter_continue=True) -> 'FeatureQuantiles':
        df = pd.read_csv(csv_path, index_col='sessID')
        return cls(df, filter_continue=filter_continue)

    @classmethod
    def fromJSON(cls, quantile_json_path: str) -> 'FeatureQuantiles':
        return cls(quantile_json_path)

    def get_quantile(self, feat: str, value, verbose: bool = False,
                     lo_to_hi: bool = True) -> int:
        quantile = bisect_left(self._quantiles[feat], value)
        if verbose:
            compare_str = "higher" if lo_to_hi else "lower"
            # print(quantile, len(self._quantiles[feat]))
            high_quant = self._quantiles[feat][quantile] if quantile < len(self._quantiles[feat]) else None
            low_quant = self._quantiles[feat][quantile - 1] if quantile > 0 else None
            quantile = quantile if lo_to_hi else 100 - quantile
            low_quant_offset = -1 if lo_to_hi else +1
            quant_low_str = f'{quantile + low_quant_offset}%={low_quant}'
            quant_high_str = f'{quantile}%={high_quant}'
            quant_str = f"{quant_low_str} and {quant_high_str}" if lo_to_hi else f"{quant_high_str} and {quant_low_str}"
            Logger.Log(
                f'A {feat} of {value} units is {compare_str} than {quantile}% (between {quant_str}) of sessions.')
        return quantile

    def _export_quantiles(self, path):
        with open(path, 'w+') as f:
            json.dump(self._quantiles, f, indent=4)



## @class FeatSeqPercentModel
# Feature to output the percentile that a session is at along a progression of gameplay timestamp features. Depends
# heavily on the logic used to produce the quantiles file. The following exemplifies current implementation:
#
# Example: A game has 10 checkpoints. Joey has reached 3/10 checkpoints. His total playtime so far is 45 seconds.
# The model will return the percentile of reaching checkpoint 4 at 45 seconds. If historically 70% of students take
# longer than 45 seconds to reach checkpoint 4, the model will return 30%.
#
# Example: A game has 10 checkpoints. Joey has reached 10/10 checkpoints. His total playtime so far is 190 seconds.
# He reached checkpoint 10 at 185 seconds. The model will return the percentile of reaching checkpoint 10 at 185
# seconds, and will continue returning that same value for this session.
#
# @param feature_sequence: sequence of timedelta features that are to be in strictly ascending order
# @param levels: levels applicable to this model
# @param time_feat: "sessDuration" or "sess_time_active", depending if the model should be used for active time or
# overall time. Make sure that the quantiles file uses the same time feature.
# @param quantile_json_path: Path to a quantiles JSON file constructed by the _FeatureQuantiles private class

class FeatSeqPercentModel(FeatureModel):
    def __init__(self, feature_sequence: List[str], levels: List[int] = [], time_feat: str = 'sess_time_active',
                 quantile_json_path: str = "models/lakeland_data/quantiles_no_continue.json"):
        self._quantile_json_path = quantile_json_path
        self._feature_sequence = feature_sequence
        self._time_feat = time_feat
        self._featureQuantiles = _FeatureQuantiles.fromJSON(
            quantile_json_path=quantile_json_path)

        super().__init__()

    def _eval(self, sess: dict, verbose: bool = False) -> Optional[float]:
        if sess['continue'] or sess['debug']:
            return None
        time_to_vals = [sess[f] for f in self._feature_sequence]
        idx_next = sum(bool(v) for v in time_to_vals)
        if idx_next < len(self._feature_sequence):
            next_feat = self._feature_sequence[idx_next]
            cur_time = sess[self._time_feat]
        else:
            next_feat = self._feature_sequence[-1]
            cur_time = time_to_vals[-1]
        if type(cur_time) is timedelta:  # sessions features give float, but cgi might give timedelta
            cur_time = cur_time.total_seconds()

        percentile_if_next_feat_now = self._featureQuantiles.get_quantile(next_feat, cur_time, verbose=verbose)

        return percentile_if_next_feat_now

    def __repr__(self):
        return f"FeatSeqPercentModel(feature_sequence={self._feature_sequence}, time_feat='{self._time_feat}'" \
               f"quantile_json_path='{self._quantile_json_path}'" \
               f", levels={self._levels}, input_type={self._input_type})"





from models.FeatureModel import FeatureModel
import typing

# This is a model that returns a dictionary with the number of homes, farms, and dairy
# farms a user bought in level 0 of their session

class TownCompositionFeatureModel(FeatureModel):
  def _eval(self, row: typing.Dict) -> dict:
      homes = row["lvl0_count_buy_home"]
      farms = row["lvl0_count_buy_farm"]
      dairy = row["lvl0_count_buy_livestock"]

      amounts = {'homes': homes, 'farms': farms, 'dairy': dairy}

      return amounts


# import standard libraries
import abc
import logging
from typing import Any, Dict, List, Set
# import local files
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class ExtractorSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._enabled     : Set[ExtractionMode]
        self._type_name   : str
        self._description : str

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For {name} Extractor config, all_elements was not a dict, defaulting to empty dict", logging.WARN)

        if "type" in all_elements.keys():
            self._type_name = ExtractorSchema._parseType(all_elements['type'])
        else:
            self._type_name = name
        if "enabled" in all_elements.keys():
            self._enabled = ExtractorSchema._parseEnabled(all_elements['enabled'])
        else:
            self._enabled = {ExtractionMode.DETECTOR, ExtractionMode.SESSION, ExtractionMode.PLAYER, ExtractionMode.POPULATION}
            Logger.Log(f"{name} config does not have an 'enabled' element; defaulting to enabled=True", logging.WARN)
        if "description" in all_elements.keys():
            self._description = ExtractorSchema._parseDescription(all_elements['description'])
        else:
            self._description = "No Description"
            Logger.Log(f"{name} config does not have an 'description' element; defaulting to description='{self._description}'", logging.WARN)

        _leftovers = { key : val for key,val in all_elements.items() if key not in {"type", "enabled", "description"} }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def TypeName(self) -> str:
        return self._type_name

    @property
    def Enabled(self) -> Set[ExtractionMode]:
        return self._enabled

    @property
    def Description(self) -> str:
        return self._description
    
    @staticmethod
    def _parseType(extractor_type):
        ret_val : str
        if isinstance(extractor_type, str):
            ret_val = extractor_type
        else:
            ret_val = str(extractor_type)
            Logger.Log(f"Extractor type was not a string, defaulting to str(type) == {ret_val}", logging.WARN)
        return ret_val

    @staticmethod
    def _parseEnabled(enabled):
        ret_val : Set[ExtractionMode] = set()
        if isinstance(enabled, bool):
            if enabled:
                ret_val = {ExtractionMode.DETECTOR, ExtractionMode.SESSION, ExtractionMode.PLAYER, ExtractionMode.POPULATION}
            else:
                ret_val = set()
        elif isinstance(enabled, list):
            for mode in enabled:
                mode = str(mode).upper()
                match mode:
                    case "DETECTOR":
                        ret_val.add(ExtractionMode.DETECTOR)
                    case "SESSION":
                        ret_val.add(ExtractionMode.SESSION)
                    case "PLAYER":
                        ret_val.add(ExtractionMode.PLAYER)
                    case "POPULATION":
                        ret_val.add(ExtractionMode.POPULATION)
                    case _:
                        Logger.Log(f"Found unrecognized element of 'enabled': {mode}", logging.WARN)
        else:
            ret_val = {ExtractionMode.DETECTOR, ExtractionMode.SESSION, ExtractionMode.PLAYER, ExtractionMode.POPULATION}
            Logger.Log(f"'enabled' element has unrecognized type {type(enabled)}; defaulting to enable all modes", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseDescription(description):
        ret_val : str
        if isinstance(description, str):
            ret_val = description
        else:
            ret_val = str(description)
            Logger.Log(f"Extractor description was not a string, defaulting to str(description) == {ret_val}", logging.WARN)
        return ret_val



# import standard libraries
from enum import IntEnum

class ExtractionMode(IntEnum):
    """Simple enum to represent the different levels of granularity at which extractions can be carried out:
    Session, Player, Population, or Detector.
    """
    SESSION = 1
    PLAYER = 2
    POPULATION = 3
    DETECTOR = 4

    def __str__(self):
        return self.name



# import standard libraries
import abc
from multiprocessing.sharedctypes import Value
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.Event import Event

## @class PerLevelFeature
class PerCountFeature(Feature):
    """PerLevelFeature
    Abstract base class for per-level game features.
    Works like a normal Feature, but checks if the given event has right "level"
    before attempting to extract from event.

    Args:
        Feature (_type_): PerLevelFeature is a subclass of Feature

    Returns:
        _type_: _description_
    """

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _validateEventCountIndex(self, event:Event):
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    @property
    def CountIndex(self) -> int:
        _ci = super().CountIndex
        if _ci is None:
            raise ValueError(f"PerCountFeature of type {type(self)} was given a null CountIndex!")
        else:
            return _ci

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _validateEvent(self, event:Event):
       
        return (
            self._validateVersion(event.LogVersion)
        and self._validateEventType(event_type=event.EventName)
        and self._validateEventCountIndex(event=event)
        )



## import standard libraries
import itertools
import logging
from datetime import datetime
from typing import Dict, List, Type, Optional, Set, Tuple, Union
## import local files
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.processors.ExtractorProcessor import ExtractorProcessor
from ogd.core.processors.PopulationProcessor import PopulationProcessor
from ogd.core.processors.PlayerProcessor import PlayerProcessor
from ogd.core.processors.SessionProcessor import SessionProcessor
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.Event import Event
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

class FeatureManager:
    def __init__(self, game_schema:GameSchema, LoaderClass:Optional[Type[GeneratorLoader]], feature_overrides:Optional[List[str]]):
        self._LoaderClass    : Optional[Type[GeneratorLoader]] = LoaderClass
        self._game_schema    : GameSchema                 = game_schema
        self._overrides      : Optional[List[str]]        = feature_overrides
        # local tracking of whether we're up-to-date on getting feature values.
        self._up_to_date     : bool                       = True
        self._latest_values  : Dict[str,List[ExportRow]]  = {}
        # local tracking of whether we used null instances in our processor hierarchies or not.
        self._used_null_play : bool                       = False
        self._used_null_sess : Dict[str, bool]            = { "null" : False }
        # Feature processors. We have all regardless of export modes set, so that we can support second-order features.
        # we have one population (since it's one pop per request), which has 0 or more players, and each player ID maps to a collection of 0 or more sessions.
        self._population : Optional[PopulationProcessor]                   = None
        self._players    : Optional[Dict[str, PlayerProcessor]]            = None
        self._sessions   : Optional[Dict[str, Dict[str,SessionProcessor]]] = None
        if self._LoaderClass is not None:
            self._population = PopulationProcessor(LoaderClass=self._LoaderClass, game_schema=game_schema,
                                                feature_overrides=feature_overrides)
            # need to initialize null instances for player and session processors, so at very least we can retrieve column names.
            self._players    = {
                "null" : PlayerProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                        player_id="null", feature_overrides=self._overrides)
            }
            self._sessions   = {
                "null" : {
                    "null" : SessionProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                            player_id="null", session_id="null", feature_overrides=self._overrides)
                }
            }
        else:
            Logger.Log(f"FeatureManager did not set up any Processors, no LoaderClass was given!", logging.WARN, depth=3)

    # TODO: make this function take list of events, and do the loop over events as low in the hierarchy as possible, which technically should be faster.
    def ProcessEvent(self, event:Event) -> None:
        # 1. process at population level.
        # NOTE: removed the skipping of unrequested modes because second-order features may need feats at levels not requested for final export.
        if self._population is not None and self._players is not None and self._sessions is not None:
            self._population.ProcessEvent(event=event)
            # 2. process at player level, adding player if needed.
            _player_id = event.UserID or "null"
            if self._LoaderClass is not None and _player_id not in self._players.keys():
                self._players[_player_id] = PlayerProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                                            player_id=_player_id,          feature_overrides=self._overrides)
            if self._LoaderClass is not None and _player_id not in self._sessions.keys():
                self._sessions[_player_id] = {}
                self._used_null_sess[_player_id] = False

            self._players[_player_id].ProcessEvent(event=event)
            if _player_id == "null":
                self._used_null_play = True
            # 3. process at session level, adding session if needed.
            if self._LoaderClass is not None and event.SessionID not in self._sessions[_player_id].keys():
                self._sessions[_player_id][event.SessionID] = SessionProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                                                    player_id=_player_id,          session_id=event.SessionID,    feature_overrides=self._overrides)
            self._sessions[_player_id][event.SessionID].ProcessEvent(event=event)
            if event.SessionID == None or event.SessionID.upper() == "NULL":
                self._used_null_sess[_player_id] = True
            self._up_to_date = False

    def ProcessFeatureData(self) -> None:
        start = datetime.now()
        Logger.Log(f"Processing Feature Data...", logging.INFO, depth=3)
        # 1. Get population 1st-order data
        if self._population is not None and self._players is not None and self._sessions is not None:
            pop_data = self._population.GetFeatureData(order=1)
            # 2. Distribute population 1st-order data
            self._population.ProcessFeatureData(feature_list=pop_data)
            for player in self._players.values():
                player.ProcessFeatureData(feature_list=pop_data)
            for session_list in self._sessions.values():
                for session in session_list.values():
                    session.ProcessFeatureData(feature_list=pop_data)
            # 3. For each player, get 1st-order data
            for player_name,player in self._players.items():
                play_data = player.GetFeatureData(order=1)
                # 4. Distribute player 1st-order data
                self._population.ProcessFeatureData(feature_list=play_data)
                player.ProcessFeatureData(feature_list=play_data)
                for session in self._sessions.get(player_name, {}).values():
                    session.ProcessFeatureData(feature_list=play_data)
            # 5. For each session, get 1st-order data
            for session_list in self._sessions.values():
                for session in session_list.values():
                    sess_data = session.GetFeatureData(order=1)
                    # 6. Distribute session 1st-order data
                    self._population.ProcessFeatureData(feature_list=sess_data)
                    player = self._players.get(session._playerID, None)
                    if player is not None:
                        player.ProcessFeatureData(feature_list=sess_data)
                    session.ProcessFeatureData(feature_list=sess_data)
            Logger.Log(f"Time to process Feature Data: {datetime.now() - start}", logging.INFO, depth=3)
        else:
            Logger.Log(f"Skipped processing of FeatureData, no feature Processors available!", logging.INFO, depth=3)

    def GetFeatureValues(self, as_str:bool = False) -> Dict[str, List[ExportRow]]:
        start = datetime.now()
        self._try_update(as_str=as_str)
        Logger.Log(f"Time to retrieve all feature values: {datetime.now() - start}", logging.INFO, depth=2)
        return self._latest_values

    def GetPopulationFeatureNames(self) -> List[str]:
        return self._population.ExtractorNames if self._population is not None else []
    def GetPopulationFeatures(self, as_str:bool = False) -> List[ExportRow]:
        start = datetime.now()
        self._try_update(as_str=as_str)
        ret_val = self._latest_values.get('population', [])
        Logger.Log(f"Time to retrieve Population lines: {datetime.now() - start} to get {len(ret_val)} lines", logging.INFO, depth=2)
        return ret_val

    def GetPlayerFeatureNames(self) -> List[str]:
        return self._players["null"].ExtractorNames if self._players is not None else []
    def GetPlayerFeatures(self, as_str:bool = False) -> List[ExportRow]:
        start   : datetime = datetime.now()
        self._try_update(as_str=as_str)
        ret_val = self._latest_values.get('players', [])
        Logger.Log(f"Time to retrieve Player lines: {datetime.now() - start} to get {len(ret_val)} lines", logging.INFO, depth=2)
        return ret_val

    def GetSessionFeatureNames(self) -> List[str]:
        return self._sessions["null"]["null"].ExtractorNames if self._sessions is not None else []
    def GetSessionFeatures(self, slice_num:int, slice_count:int, as_str:bool = False) -> List[ExportRow]:
        start   : datetime = datetime.now()
        self._try_update(as_str=as_str)
        ret_val = self._latest_values.get('sessions', [])
        time_delta = datetime.now() - start
        Logger.Log(f"Time to retrieve Session lines for slice [{slice_num}/{slice_count}]: {time_delta} to get {len(ret_val)} lines", logging.INFO, depth=2)
        return ret_val

    def ClearPopulationLines(self) -> None:
        if self._population is not None:
            self._population.ClearLines()
    def ClearPlayerLines(self) -> None:
        if self._players is not None and self._LoaderClass is not None:
            for player in self._players.values():
                player.ClearLines()
            self._players = {}
            self._players["null"] = PlayerProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                                    player_id="null", feature_overrides=self._overrides)
    def ClearSessionLines(self) -> None:
        if self._sessions is not None and self._LoaderClass is not None:
            for sess_list in self._sessions.values():
                for sess in sess_list.values():
                    sess.ClearLines()
            self._sessions = {}
            self._sessions["null"] = {
                "null" : SessionProcessor(LoaderClass=self._LoaderClass, game_schema=self._game_schema,
                                        player_id="null", session_id="null", feature_overrides=self._overrides)
            }

    def _flatHierarchy(self) -> List[ExtractorProcessor]:
        ret_val : List[ExtractorProcessor] = []
        if self._population is not None:
            ret_val = [self._population]
        if self._players is not None:
            ret_val += self._players.values()
        if self._sessions is not None:
            for sess_list in self._sessions.values():
                ret_val += sess_list.values()
        return ret_val

    def _try_update(self, as_str:bool = False):
        if not self._up_to_date:
            self.ProcessFeatureData()
            # for some reason, this didn't work as sum over list of lists, so get sessions manually with a normal loop:
            list_o_playlists : List[List[ExportRow]]       = [player.Lines for player_id,player in self._players.items() if (player_id != "null" or self._used_null_play)] if self._players is not None else []
            flat_playlist    : List[ExportRow]             = list(itertools.chain.from_iterable(list_o_playlists))
            list_o_sesslists : List[List[List[ExportRow]]] = [[session.Lines for session_id,session in session_list.items() if (session_id != "null" or self._used_null_sess[player_name])] for player_name,session_list in self._sessions.items()] if self._sessions is not None else []
            flat_sesslist    : List[ExportRow]             = list(itertools.chain.from_iterable(itertools.chain.from_iterable(list_o_sesslists)))
            self._latest_values = {
                "population" : self._population.Lines if self._population is not None else [],
                "players"    : flat_playlist,
                "sessions"   : flat_sesslist
            }
            self._up_to_date = True



import csv
import datetime
import time
from operator import itemgetter
from collections import defaultdict
import winsound

class ClassroomDetector:
    def __init__(self, data_table, dt_format = "%Y-%m-%d %H:%M:%S",
                class_size_lower = 10, class_size_upper = 45, class_window = 600,
                duration_lower = 180, duration_upper = 4200, hexads = 4, afk_duration = 60):
        self.data_table = data_table
        self.dt_format  = dt_format
        self.class_size_lower = class_size_lower
        self.class_size_upper = class_size_upper
        self.class_window     = class_window
        self.duration_lower   = duration_lower
        self.duration_upper   = duration_upper
        self.hexads       = hexads
        self.afk_duration = afk_duration

    def process_main(self, _data, _dt_fmt=None):
        '''
        Shell function for handling various classroom identification operations.
        '''
        header = {v: k for k, v in enumerate(_data[0])}
        prepro_data = self.preprocess(_data=_data, _header=header)

        ip_session_dict = defaultdict(list)
        last_timestamp = prepro_data[1][header['server_time']]
        ids_to_exclude = []

        for n, i in enumerate(prepro_data):
            # if a session id changes, we need to record the new start time and old end time
            if i[header['session_id']] != prepro_data[n-1][header['session_id']]:
                old_key = prepro_data[n-1][header['remote_addr']] + "_" + prepro_data[n-1][header['session_id']]
                ip_session_dict[old_key].append(prepro_data[n-1][header['server_time']])

                new_key = i[header['remote_addr']] + "_" + i[header['session_id']]
                ip_session_dict[new_key].append(i[header['server_time']])

            # while we're within a session, if there's a long gap of inactivity, we need to exclude that session
            elif i[header['server_time']] - last_timestamp > self.afk_duration:
                ids_to_exclude.append(i[header['remote_addr']] + "_" + i[header['session_id']])

            last_timestamp = i[header['server_time']]

        # removing the afk sessions from our analysis
        ip_session_dict.pop('remote_addr_session_id')
        for i in ids_to_exclude:
            try:
                ip_session_dict.pop(i)
            except:
                pass

        overlaps_bystart = defaultdict(list)
        overlaps_byend = defaultdict(list)

        for key in ip_session_dict:
            session_start = int(ip_session_dict[key][0])
            session_end = int(ip_session_dict[key][-1])

            for key2 in ip_session_dict:
                # overlap = False
                if key2 != key:
                    if session_start <= int(ip_session_dict[key2][0]) <= session_end:
                        overlaps_bystart[key].append([key2,
                                                    session_start,
                                                    session_end,
                                                    ip_session_dict[key2][0],
                                                    ip_session_dict[key2][-1]])
                        # overlap = True
                    if session_start <= int(ip_session_dict[key2][-1]) <= session_end:
                        overlaps_byend[key].append(key2)
                        # overlap = True
        return overlaps_bystart
        

    def preprocess(self, _data, _header, _dt_fmt=None):
        '''
        Shortens remote_addr to the first three digits of session IP.
        Converts server_time to UNIX
        '''
        _dt_fmt = _dt_fmt if _dt_fmt is not None else self.dt_format
        header_list = _data[0]
        for i in _data[1:]:
            if self.hexads != 4:
                ip_shortened = i[_header['remote_addr']].split(".")[:(4 - self.hexads) * -1]
                i[_header['remote_addr']] = ".".join(ip_shortened)
            i[_header['server_time']] = int(
                time.mktime(datetime.datetime.strptime(i[_header['server_time']], _dt_fmt).timetuple()))

        # sorting the file in preparation for iterating through it
        _data = sorted(_data[1:], key=itemgetter(_header['remote_addr'], _header['session_id'], _header['server_time']))
        _data = [x for x in _data[1:] if int(x[_header['event_custom']]) <= 16] # selecting only student actions
        _data = [header_list] + _data # adding the header row back in
        return _data

if __name__ == '__main__':
    '''
    various config settings:
    file_input is input raw .csv files
    delim_type is the delimiter for the raw; usually this is \t
    dt_format is the datetime format from the "server_time" column
    class_size is the bounds on the number of unique session IDs that are considered to be a class
    class_window is the max duration between any two unique session starts to be considered a class
    duration is the bounds on the duration of a given game session
    hexads are the number of hexads in the ip that you want to look at
    afk duration is the maximum amount of time between two user-generated actions to not be considered afk
    '''
    #file_input = "D:\\Field Day Lab\\Raw_LAKELAND_2019Dec_iter1.csv"
    file_input = "D:\\Luke Swanson\\D_Documents\\!work\\FieldDay\\LAKELAND_20200301_to_20200331_a9720c1_raw.csv"
    #file_input = "D:\\Field Day Lab\\LAKELAND_20200401_to_20200430\\LAKELAND_20200401_to_20200430_a9720c1_raw.csv"
    file_output = file_input[:-4] + "_classroom_subset_new_code.csv"
    delim_type = "\t"

    with open(file_input) as f:
        in_data = list(csv.reader(f,delimiter=delim_type))
    class_detect = ClassroomDetector(in_data)
    overlaps_bystart = class_detect.process_main(in_data)

    writer = csv.writer(open("Lakeland IP Blocks By Timestamp.csv","w",newline=""))
    for key in overlaps_bystart:
        if len(overlaps_bystart[key]) > 10:
            writer.writerow(key)
            for val in overlaps_bystart[key]:
                writer.writerow(val)
            writer.writerow([])


## import standard libraries
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.WAVES.features import *
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

## @class WaveExtractor
#  Extractor subclass for extracting features from Waves game data.
class WaveLoader(GeneratorLoader):

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @staticmethod
    def _getFeaturesModule():
        return features
    
    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        # Session-level features.
        if extractor_params._count_index is None:
            match feature_type:
                case "AverageFails":
                    ret_val = AverageFails.AverageFails(params=extractor_params)
                case "AverageLevelTime":
                    ret_val = AverageLevelTime.AverageLevelTime(params=extractor_params)
                case "AverageMoveTypeChanges":
                    ret_val = AverageMoveTypeChanges.AverageMoveTypeChanges(params=extractor_params)
                case "AverageSliderMoves":
                    ret_val = AverageSliderMoves.AverageSliderMoves(params=extractor_params)
                case "OverallPercentAmplitudeMoves":
                    ret_val = OverallPercentAmplitudeMoves.OverallPercentAmplitudeMoves(params=extractor_params)
                case "OverallPercentOffsetMoves":
                    ret_val = OverallPercentOffsetMoves.OverallPercentOffsetMoves(params=extractor_params)
                case "OverallPercentWavelengthMoves":
                    ret_val = OverallPercentWavelengthMoves.OverallPercentWavelengthMoves(params=extractor_params)
                case "OverallSliderAverageRange":
                    ret_val = OverallSliderAverageRange.OverallSliderAverageRange(params=extractor_params)
                case "OverallSliderAverageStandardDeviations":
                    ret_val = OverallSliderAverageStandardDeviations.OverallSliderAverageStandardDeviations(params=extractor_params)
                case "PersistentSessionID":
                    ret_val = PersistentSessionID.PersistentSessionID(params=extractor_params)
                case "SessionID":
                    ret_val = SessionID.SessionID(params=extractor_params, sessionID=self._session_id)
                    # sequence features
                case "SequenceLevel":
                    ret_val = SequenceLevel.SequenceLevel(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid aggregate feature for Waves.")
        # Per-count features
        # level attempt features
        else:
            match feature_type:
                case "BeginCount":
                    ret_val = BeginCount.BeginCount(params=extractor_params)
                case "Completed":
                    ret_val = Completed.Completed(params=extractor_params)
                case "MenuButtonCount":
                    ret_val = MenuButtonCount.MenuButtonCount(params=extractor_params)
                case "SucceedCount":
                    ret_val = SucceedCount.SucceedCount(params=extractor_params)
                case "TotalFails":
                    ret_val = TotalFails.TotalFails(params=extractor_params)
                case "TotalLevelTime":
                    ret_val = TotalLevelTime.TotalLevelTime(params=extractor_params)
                case "TotalResets":
                    ret_val = TotalResets.TotalResets(params=extractor_params)
                case "TotalSkips":
                    ret_val = TotalSkips.TotalSkips(params=extractor_params)
                    # slider move features
                case "FirstMoveType":
                    ret_val = FirstMoveType.FirstMoveType(params=extractor_params)
                case "AmplitudeGoodMoveCount":
                    ret_val = AmplitudeGoodMoveCount.AmplitudeGoodMoveCount(params=extractor_params)
                case "OffsetGoodMoveCount":
                    ret_val = OffsetGoodMoveCount.OffsetGoodMoveCount(params=extractor_params)
                case "WavelengthGoodMoveCount":
                    ret_val = WavelengthGoodMoveCount.WavelengthGoodMoveCount(params=extractor_params)
                case "PercentAmplitudeMoves":
                    ret_val = PercentAmplitudeMoves.PercentAmplitudeMoves(params=extractor_params)
                case "PercentOffsetMoves":
                    ret_val = PercentOffsetMoves.PercentOffsetMoves(params=extractor_params)
                case "PercentWavelengthMoves":
                    ret_val = PercentWavelengthMoves.PercentWavelengthMoves(params=extractor_params)
                case "PercentAmplitudeGoodMoves":
                    ret_val = PercentAmplitudeGoodMoves.PercentAmplitudeGoodMoves(params=extractor_params)
                case "PercentOffsetGoodMoves":
                    ret_val = PercentOffsetGoodMoves.PercentOffsetGoodMoves(params=extractor_params)
                case "PercentWavelengthGoodMoves":
                    ret_val = PercentWavelengthGoodMoves.PercentWavelengthGoodMoves(params=extractor_params)
                case "SliderAverageRange":
                    ret_val = SliderAverageRange.SliderAverageRange(params=extractor_params)
                case "SliderAverageStandardDeviations":
                    ret_val = SliderAverageStandardDeviations.SliderAverageStandardDeviations(params=extractor_params)
                case "TotalArrowMoves":
                    ret_val = TotalArrowMoves.TotalArrowMoves(params=extractor_params)
                case "TotalMoveTypeChanges":
                    ret_val = TotalMoveTypeChanges.TotalMoveTypeChanges(params=extractor_params)
                case "TotalSliderMoves":
                    ret_val = TotalSliderMoves.TotalSliderMoves(params=extractor_params)
                    # level progress over time features
                # case "ClosenessIntercept":
                #     ret_val = ClosenessIntercept.ClosenessIntercept(params=extractor_params)
                # case "ClosenessR2":
                #     ret_val = ClosenessR2.ClosenessR2(params=extractor_params)
                # case "ClosenessSlope":
                #     ret_val = ClosenessSlope.ClosenessSlope(params=extractor_params)
                # case "RangeIntercept":
                #     ret_val = RangeIntercept.RangeIntercept(params=extractor_params)
                # case "RangeR2":
                #     ret_val = RangeR2.RangeR2(params=extractor_params)
                # case "RangeSlope":
                #     ret_val = RangeSlope.RangeSlope(params=extractor_params)
                    # quiz features
                case "QuestionAnswered":
                    ret_val = QuestionAnswered.QuestionAnswered(params=extractor_params)
                case "QuestionCorrect":
                    ret_val = QuestionCorrect.QuestionCorrect(params=extractor_params)
                case "TimeToAnswerMS":
                    ret_val = TimeToAnswerMS.TimeToAnswerMS(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid per-count feature for Waves.")
        return ret_val

    def _loadDetector(self, detector_type:str, name:str, detector_args:Dict[str,Any], trigger_callback:Callable[[Event], None], count_index:Optional[int] = None) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid detector for Waves.")

    # *** BUILT-INS & PROPERTIES ***

    ## Constructor for the WaveLoader class.
    def __init__(self, player_id:str, session_id: str, game_schema:GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]=None):
        """Constructor for the WaveLoader class.

        :param player_id: _description_
        :type player_id: str
        :param session_id: The id number for the session whose data is being processed by this instance
        :type session_id: str
        :param game_schema: A data structure containing information on how the game events and other data are structured
        :type game_schema: GameSchema
        :param feature_overrides: A list of features to export, overriding the default of exporting all enabled features.
        :type feature_overrides: Optional[List[str]]
        """
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)
        # self._last_adjust_type : Optional[str] = None
        # self.start_times: Dict       = {}
        # self.end_times:   Dict       = {}
        # self.amp_move_counts:  Dict   = {}
        # self.off_move_counts:  Dict   = {}
        # self.wave_move_counts: Dict   = {}
        # self.saw_first_move: Dict[int, bool] = {}
        # self.latest_complete_lvl8 = None
        # self.latest_complete_lvl16 = None
        # self.latest_answer_Q0 = None
        # self.latest_answer_Q2 = None
        # self.active_begin = None
        # self.move_closenesses_tx: Dict = {}
        # self._features.setValByName(feature_name="sessionID", new_value=session_id)
        # # we specifically want to set the default value for questionAnswered to None, for unanswered.
        # for ans in self._features.getValByName(feature_name="questionAnswered").keys():
        #     self._features.setValByIndex(feature_name="questionAnswered", index=ans, new_value=None)
        # for q in self._features.getValByName(feature_name="questionCorrect"):
        #     self._features.setValByIndex(feature_name="questionCorrect", index=q, new_value=None)
        # for elem in self._features.getValByName(feature_name="firstMoveType"):
        #     self._features.setValByIndex(feature_name="firstMoveType", index=elem, new_value=None)

    ## Function to perform extraction of features from a row.
    #
    #  @param row_with_complex_parsed A row of game data from the db, with the
    #                                 "complex data" already parsed from JSON.
    #  @param game_table  A data structure containing information on how the db
    #                     table assiciated with this game is structured.
    # def _extractFeaturesFromEvent(self, event:Event, table_schema:TableSchema):
    #     if event.SessionID == self._session_id:
    #         level = event.GameState['level']
    #         # If we haven't set persistent id, set now.
    #         if self._features.getValByName(feature_name="persistentSessionID") == 0:
    #             self._features.setValByName(feature_name="persistentSessionID", new_value=event.UserData['persistent_session_id'])
    #         # Ensure we have private data initialized for the given level.
    #         if not level in self._levels:
    #             bisect.insort(self._levels, level)
    #             self._features.initLevel(level)
    #             self.amp_move_counts[level] = 0
    #             self.off_move_counts[level] = 0
    #             self.wave_move_counts[level] = 0
    #         # Then, handle cases for each type of event
    #         # NOTE: for BEGIN and COMPLETE, we assume only one event of each type happens.
    #         # If there are somehow multiples, the previous times are overwritten by the newer ones.
    #         # 1) figure out what type of event we had. If CUSTOM, we'll use the event_custom sub-item.
    #         event_type = event.EventName.split('.')[0]
    #         if event_type == "CUSTOM":
    #             event_type = event.EventData['event_custom']
    #         # 2) handle cases for each type of event
    #         match event_type:
    #             case "BEGIN":
    #                 self._extractFromBegin(level=level, event_client_time=event.Timestamp)
    #             case "COMPLETE":
    #                 self._extractFromComplete(level=level, event_client_time=event.Timestamp)
    #             case "SUCCEED":
    #                 self._extractFromSucceed(level=level)
    #             case "MENU_BUTTON":
    #                 self._extractFromMenuBtn(level=level, event_client_time=event.Timestamp)
    #             case "SKIP_BUTTON":
    #                 self._extractFromSkipBtn(level=level)
    #             case "DISMISS_MENU_BUTTON":
    #                 pass
    #                 # print("Stub: Got a DISMISS_MENU_BUTTON event, nothing to do with it.")
    #             case "RESET_BTN_PRESS":
    #                 self._extractFromResetBtnPress(level=level)
    #             case "FAIL":
    #                 self._extractFromFail(level=level)
    #             case ["SLIDER_MOVE_RELEASE", "ARROW_MOVE_RELEASE"] :
    #                 self._extractFromMoveRelease(level=level, event_data=event.EventData, event_client_time=event.Timestamp)
    #             case "QUESTION_ANSWER":
    #                 self._extractFromQuestionAnswer(event_data=event.EventData, event_client_time=event.Timestamp)
    #                 # print("Q+A: " + str(event_data))
    #             case _:
    #                 raise Exception(f"Found an unrecognized event type: {event_type}")
    #     else:
    #         Logger.Log(f"Got a row with incorrect session id! Expected {self._session_id}, got {event.SessionID}!", logging.ERROR)
                                               
    ## Function to perform calculation of aggregate features from existing
    #  per-level/per-custom-count features.
    # def CalculateAggregateFeatures(self):
    #     if len(self._levels) > 0:
    #         # Calculate per-level averages and percentages, since we can't calculate
    #         # them until we know how many total events occur.
    #         for lvl in self._levels:
    #             total_slider_moves = self._features.getValByIndex(feature_name="totalSliderMoves", index=lvl)
    #             total_moves = total_slider_moves + self._features.getValByIndex(feature_name="totalArrowMoves", index=lvl)
    #             # percents of each move type
    #             try:
    #                 val   = self.amp_move_counts[lvl] / total_moves if total_moves > 0 else total_moves
    #             except Exception as err:
    #                 val = None
    #                 Logger.Log(f"Currently, total_moves = {total_moves}")
    #                 msg = f"{type(err)} {str(err)}"
    #                 Logger.Log(msg, logging.ERROR)
    #                 traceback.print_tb(err.__traceback__)
    #             self._features.setValByIndex(feature_name="percentAmplitudeMoves", index=lvl, new_value=val)
    #             val   = self.off_move_counts[lvl] / total_moves if total_moves > 0 else total_moves
    #             self._features.setValByIndex(feature_name="percentOffsetMoves", index=lvl, new_value=val)
    #             val   = self.wave_move_counts[lvl] / total_moves if total_moves > 0 else total_moves
    #             self._features.setValByIndex(feature_name="percentWavelengthMoves", index=lvl, new_value=val)
    #             # percents of good moves for each type
    #             val   = self._features.getValByIndex(feature_name="amplitudeGoodMoveCount", index=lvl) / self.amp_move_counts[lvl] \
    #                  if lvl in self.amp_move_counts.keys() and self.amp_move_counts[lvl] > 0 else 0.0
    #             self._features.setValByIndex(feature_name="percentAmplitudeGoodMoves", index=lvl, new_value=val)
    #             val   = self._features.getValByIndex(feature_name="offsetGoodMoveCount", index=lvl) / self.off_move_counts[lvl] \
    #                  if lvl in self.off_move_counts.keys() and self.off_move_counts[lvl] > 0 else 0.0
    #             self._features.setValByIndex(feature_name="percentOffsetGoodMoves", index=lvl, new_value=val)
    #             val   = self._features.getValByIndex(feature_name="wavelengthGoodMoveCount", index=lvl) / self.wave_move_counts[lvl] \
    #                  if lvl in self.wave_move_counts.keys() and self.wave_move_counts[lvl] > 0 else 0.0
    #             self._features.setValByIndex(feature_name="percentWavelengthGoodMoves", index=lvl, new_value=val)
    #             # avg slider std devs and ranges.
    #             total_slider_stdevs = self._features.getValByIndex(feature_name="sliderAvgStdDevs", index=lvl)
    #             val = total_slider_stdevs / total_slider_moves if total_slider_moves > 0 else total_slider_moves
    #             self._features.setValByIndex(feature_name="sliderAvgStdDevs", index=lvl, new_value=val)
    #             total_slider_range  = self._features.getValByIndex(feature_name="sliderAvgRange", index=lvl)
    #             val = total_slider_range / total_slider_moves if total_slider_moves > 0 else total_slider_moves
    #             self._features.setValByIndex(feature_name="sliderAvgRange", index=lvl, new_value=val)
    #         # Then, calculate true aggregates.
    #         num_lvl = len(self._levels)
    #         all_vals = [elem["val"] for elem in self._features.getValByName(feature_name="totalSliderMoves").values() if elem["val"] is not None]
    #         self._features.setValByName(feature_name="avgSliderMoves", new_value=sum(all_vals) / num_lvl)

    #         all_vals = [elem["val"] for elem in self._features.getValByName(feature_name="totalLevelTime").values() if elem["val"] is not None]
    #         self._features.setValByName(feature_name="avgLevelTime", new_value=sum(all_vals) / num_lvl)

    #         all_vals = [elem["val"] for elem in self._features.getValByName(feature_name="totalMoveTypeChanges").values() if elem["val"] is not None]
    #         self._features.setValByName(feature_name="avgMoveTypeChanges", new_value=sum(all_vals) / num_lvl)

    #         # Determine number of moves that occurred across all levels
    #         all_amps = sum(list(self.amp_move_counts.values()))
    #         all_offs = sum(list(self.off_move_counts.values()))
    #         all_waves = sum(list(self.wave_move_counts.values()))
    #         all_moves = all_amps + all_offs + all_waves

    #         # Calculate percentages of each move type over all moves.
    #         perc_amps = all_amps / all_moves if all_moves > 0 else all_moves
    #         perc_offs = all_offs / all_moves if all_moves > 0 else all_moves
    #         perc_waves = all_waves / all_moves if all_moves > 0 else all_moves
    #         self._features.setValByName(feature_name="overallPercentAmplitudeMoves", new_value=perc_amps)
    #         self._features.setValByName(feature_name="overallPercentOffsetMoves", new_value=perc_offs)
    #         self._features.setValByName(feature_name="overallPercentWavelengthMoves", new_value=perc_waves)

    #         # Calculate average ranges and std devs over all moves.
    #         all_stdevs = sum([elem["val"] for elem in self._features.getValByName(feature_name="sliderAvgStdDevs").values() if elem["val"] is not None])
    #         all_ranges = sum([elem["val"] for elem in self._features.getValByName(feature_name="sliderAvgRange").values() if elem["val"] is not None])
    #         avg_stdevs = all_stdevs / all_moves if all_moves > 0 else all_moves
    #         avg_ranges = all_ranges / all_moves if all_moves > 0 else all_moves
    #         self._features.setValByName(feature_name="overallSliderAvgStdDevs", new_value=avg_stdevs)
    #         self._features.setValByName(feature_name="overallSliderAvgRange", new_value=avg_ranges)

    #         # Finally, calculate average fails per level
    #         all_fails = sum([elem["val"] for elem in self._features.getValByName(feature_name="totalFails").values() if elem["val"] is not None])
    #         avg_fails = all_fails / all_moves if all_moves > 0 else all_moves
    #         self._features.setValByName(feature_name="avgFails", new_value=avg_fails)

    ## Private function to extract features from a "BEGIN" event.
    #  The features affected are:
    #  - start_times (used to calculate totalLevelTime and avgLevelTime)
    #  - beginCount
    #  - totalLevelTime
    #
    #  @param level             The level being played when event occurred.
    #  @param event_client_time The time when this event occurred, according to game client.
    # def _extractFromBegin(self, level:int, event_client_time:datetime):
    #     self._features.incValByIndex(feature_name="beginCount", index=level)
    #     if self.active_begin == None:
    #         self.start_times[level] = event_client_time
    #         self.move_closenesses_tx[level] = {'t': [], 'completeness': [], 'range': []}
    #     elif self.active_begin == level:
    #         pass # in this case, just keep going.
    #     else:
    #         self.end_times[self.active_begin] = event_client_time
    #         self._calc_level_end(self.active_begin)
    #         try:
    #             self._features.incValByIndex(feature_name="totalLevelTime", index=self.active_begin, increment=self._calcLevelTime(self.active_begin))
    #         except Exception as err:
    #             raise(Exception(f"{type(err)} {str(err)}, level={level}, method=extractFromBegin, active_begin={self.active_begin}"))
    #         self.start_times[level] = event_client_time
    #         self.move_closenesses_tx[level] = {'t': [], 'completeness': [], 'range': []}
    #     # in any case, current level now has active begin event.
    #     self.active_begin = level

    # ## Private function to extract features from a "COMPLETE" event.
    # #  The features affected are:
    # #  - end_times (used to calculate totalLevelTime and avgLevelTime)
    # #  - completed
    # #  - completeCount
    # #  - totalLevelTime
    # #
    # #  @param level             The level being played when event occurred.
    # #  @param event_client_time The time when this event occurred, according to game client.
    # def _extractFromComplete(self, level:int, event_client_time:datetime):
    #     self._features.incValByIndex(feature_name="completeCount", index=level)
    #     # track latest completion of levels 8 & 16, for the timeToAnswerMS features.
    #     if (level == 8):
    #         self.latest_complete_lvl8 = event_client_time
    #     elif (level == 16):
    #         self.latest_complete_lvl16 = event_client_time
    #     # Handle tracking of level play times.
    #     if self.active_begin == None:
    #         sess_id = self._features.getValByName(feature_name="sessionID")
    #         Logger.Log(f"Got a 'Complete' event when there was no active 'Begin' event! Sess ID: {sess_id}, level: {level}", logging.ERROR)
    #     elif self.active_begin != level:
    #         sess_id = self._features.getValByName(feature_name="sessionID")
    #         Logger.Log(f"Got a 'Complete' event when the active 'Begin' was for a different level ({self.active_begin})! Sess ID: {sess_id}, level: {level}", logging.ERROR)
    #     else:
    #         self.end_times[level] = event_client_time
    #         self._calc_level_end(level)
    #         self._features.setValByIndex(feature_name="completed", index=level, new_value=1)
    #         try:
    #             self._features.incValByIndex(feature_name="totalLevelTime", index=level, increment=self._calcLevelTime(level))
    #         except Exception as err:
    #             raise(Exception(f"{type(err)} {str(err)}, level={level}, method=extractFromComplete, active_begin={self.active_begin}"))
    #         self.active_begin = None

    # ## Private function to extract features from a "SUCCEED" event.
    # #  The features affected are:
    # #  - succeedCount
    # #
    # #  @param level The level being played when succeed event occurred.
    # def _extractFromSucceed(self, level:int):
    #     self._features.incValByIndex(feature_name="succeedCount", index=level)

    # ## Private function to extract features from a "MENU_BUTTON" event.
    # #  The features affected are:
    # #  - end_times (used to calculate totalLevelTime and avgLevelTime)
    # #  - menuBtnCount
    # #  - totalLevelTime
    # #
    # #  @param level             The level being played when event occurred.
    # #  @param event_client_time The time when this event occurred, according to game client.
    # def _extractFromMenuBtn(self, level:int, event_client_time:datetime):
    #     self._features.incValByIndex(feature_name="menuBtnCount", index=level)
    #     if self.active_begin == None:
    #         sess_id = self._features.getValByName(feature_name="sessionID")
    #         Logger.Log(f"Got a 'Menu Button' event when there was no active 'Begin' event! Sess ID: {sess_id}, level: {level}", logging.ERROR)
    #     elif self.active_begin != level:
    #         sess_id = self._features.getValByName(feature_name="sessionID")
    #         Logger.Log(f"Got a 'Menu Button' event when the active 'Begin' was for a different level ({self.active_begin})! Sess ID: {sess_id}, level: {level}", logging.ERROR)
    #     else:
    #         self.end_times[level] = event_client_time
    #         self._calc_level_end(level)
    #         try:
    #             self._features.incValByIndex(feature_name="totalLevelTime", index=level, increment=self._calcLevelTime(level))
    #         except Exception as err:
    #             raise(Exception(f"{type(err)} {str(err)}, level={level}, method=extractFromMenuBtn, active_begin={self.active_begin}"))
    #         self.active_begin = None

    # ## Private function to extract features from a "SKIP_BUTTON" event.
    # #  The features affected are:
    # #  - totalSkips
    # #
    # #  @param level The level being played when reset button was pressed.
    # def _extractFromSkipBtn(self, level:int):
    #     self._features.incValByIndex(feature_name="totalSkips", index=level)
    #     self.active_begin = None

    # ## Private function to extract features from a "RESET_BTN_PRESS" event.
    # #  The features affected are:
    # #  - totalResets
    # #
    # #  @param level The level being played when reset button was pressed.
    # def _extractFromResetBtnPress(self, level:int):
    #     self._features.incValByIndex(feature_name="totalResets", index=level)

    # ## Private function to extract features from a "FAIL" event.
    # #  The features affected are:
    # #  - totalFails
    # #
    # #  @param level The level being played when the event occurred.
    # def _extractFromFail(self, level:int):
    #     self._features.incValByIndex(feature_name="totalFails", index=level)

    # ## Private function to extract features from a "SLIDER_MOVE_RELEASE" or 
    # #  "ARROW_MOVE_RELEASE" event.
    # #  The features affected are:
    # #  - totalMoveTypeChanges
    # #  - totalSliderMoves
    # #  - totalArrowMoves
    # #  - sliderAvgStdDevs
    # #  - sliderAvgRange
    # #
    # #  @param level      The level being played when event occurred.
    # #  @param event_data Parsed JSON data from the row being processed.
    # def _extractFromMoveRelease(self, level:int, event_data:Dict[str,Any], event_client_time:datetime):
    #     def isGoodMove(event_data):
    #         end_distance   = abs(event_data["correct_val"] - event_data["end_val"])
    #         start_distance = abs(event_data["correct_val"] - event_data["begin_val"])
    #         return end_distance < start_distance
    #     # first, log the type of move.
    #     if event_data["slider"] != self._last_adjust_type:
    #         self._last_adjust_type = event_data["slider"]
    #         # NOTE: We assume data are processed in order of event time.
    #         # If events are not sorted by time, the "move type changes" may be inaccurate.
    #         self._features.incValByIndex(feature_name="totalMoveTypeChanges", index=level)
    #     if event_data["slider"] == "AMPLITUDE":
    #         self.amp_move_counts[level] += 1
    #         if isGoodMove(event_data):
    #             self._features.incValByIndex(feature_name="amplitudeGoodMoveCount", index=level)
    #         if not level in self.saw_first_move.keys() or not self.saw_first_move[level]:
    #             self.saw_first_move[level] = True
    #             self._features.setValByIndex(feature_name="firstMoveType", index=level, new_value='A')
    #     elif event_data["slider"] == "OFFSET":
    #         self.off_move_counts[level] += 1
    #         if isGoodMove(event_data):
    #             self._features.incValByIndex(feature_name="offsetGoodMoveCount", index=level)
    #         if not level in self.saw_first_move.keys() or not self.saw_first_move[level]:
    #             self.saw_first_move[level] = True
    #             self._features.setValByIndex(feature_name="firstMoveType", index=level, new_value='O')
    #     elif event_data["slider"] == "WAVELENGTH":
    #         self.wave_move_counts[level] += 1
    #         if isGoodMove(event_data):
    #             self._features.incValByIndex(feature_name="wavelengthGoodMoveCount", index=level)
    #         if not level in self.saw_first_move.keys() or not self.saw_first_move[level]:
    #             self.saw_first_move[level] = True
    #             self._features.setValByIndex(feature_name="firstMoveType", index=level, new_value='W')
    #     # then, log things specific to slider move:
    #     if event_data["event_custom"] == "SLIDER_MOVE_RELEASE":
    #         self._features.incValByIndex(feature_name="totalSliderMoves", index=level)
    #         # These will be averages over the level. Per-row, just accumulate,
    #         # then we'll divide in the aggregate feature step.
    #         self._features.incValByIndex(feature_name="sliderAvgStdDevs", index=level,
    #                                     increment=event_data["stdev_val"])
    #         self._features.incValByIndex(feature_name="sliderAvgRange", index=level,
    #                                     increment= event_data["max_val"]
    #                                              - event_data["min_val"])
    #     else: # log things specific to arrow move:
    #         self._features.incValByIndex(feature_name="totalArrowMoves", index=level)

    #     if event_data['event_custom'] == 'ARROW_MOVE_RELEASE':
    #         begin_closeness, end_closeness = None, event_data['closeness']
    #         range = event_data['begin_val'] - event_data['end_val']
    #         range = range if range > 0 else range*-1
    #     else: #SLIDER MOVE RELEASE
    #         begin_closeness, end_closeness = event_data['begin_closeness'], event_data['end_closeness']
    #         range = event_data['max_val'] - event_data['min_val']

    #     if begin_closeness and not self.move_closenesses_tx[level]:
    #         self.move_closenesses_tx[level]['completeness'].append(begin_closeness)
    #         self.move_closenesses_tx[level]['t'].append(self.start_times[level])
    #         self.move_closenesses_tx[level]['range'].append(None)
    #     self.move_closenesses_tx[level]['completeness'].append(end_closeness)
    #     self.move_closenesses_tx[level]['t'].append(event_client_time)
    #     self.move_closenesses_tx[level]['range'].append(range)




    # ## Private function to extract features from a "QUESTION_ANSWER" event.
    # #  The features affected are:
    # #  - questionAnswered
    # #  - questionCorrect
    # #
    # #  @param event_data Parsed JSON data from the row being processed.
    # def _extractFromQuestionAnswer(self, event_data:Dict[str,Any], event_client_time:datetime):
    #     q_num = event_data["question"]
    #     if (q_num == 0):
    #         self.latest_answer_Q0 = event_client_time
    #     elif (q_num == 2):
    #         self.latest_answer_Q2 = event_client_time
    #     answer_time = self._calcAnswerTime(q_num=q_num, event_client_time=event_client_time)
    #     self._features.setValByIndex(feature_name="timeToAnswerMS", index=q_num, new_value=answer_time)
    #     self._features.setValByIndex(feature_name="questionAnswered", index=q_num,
    #                                 new_value=event_data["answered"])
    #     correctness = 1 if event_data["answered"] == event_data["answer"] else 0
    #     self._features.setValByIndex(feature_name="questionCorrect", index=q_num, new_value=correctness)

    # ## Private function to calculate percent of moves made of a given type,
    # #  for a given level.
    # #
    # #  @param key   The feature name for the type of move to calculate percentage on.
    # #  @param level The level for which we want to calculate a percentage.
    # def _calcPercentMoves(self, key:str, level:int) -> float:
    #     all_vals = [elem["val"] for elem in self._features.getValByName(feature_name=key).values()]
    #     num = sum(all_vals)
    #     if num == 0:
    #         return 0
    #     else:
    #         slider_vals = [elem["val"] for elem in self._features.getValByName(feature_name="totalSliderMoves").values()]
    #         arrow_vals = [elem["val"] for elem in self._features.getValByName(feature_name="totalArrowMoves").values()]
    #         denom = sum(slider_vals) + sum(arrow_vals)
    #         return num/denom

    # ## Private function to calculate the time spent on a given level.
    # #
    # #  @param lvl      The level whose play time should be calculated
    # #  @return         The play time if level is valid, -1 if level
    # #                  was not completed, 0 if level was never attempted.
    # def _calcLevelTime(self, lvl:int) -> int:
    #     # use 0 if not played, -1 if not completed
    #     if lvl in self._levels:
    #         if lvl in self.start_times and lvl in self.end_times:
    #             return (self.end_times[lvl] - self.start_times[lvl]).total_seconds()
    #         else:
    #             raise Exception(f"Level {lvl} is missing a start or end time")
    #     else:
    #         return 0
        
    # def _calcAnswerTime(self, q_num:int, event_client_time) -> Optional[int]:
    #     millis: Optional[float]
    #     if q_num == 0:
    #         millis = 1000.0 * (event_client_time - self.latest_complete_lvl8).total_seconds()
    #     elif q_num == 1:
    #         millis = 1000.0 * (event_client_time - self.latest_answer_Q0).total_seconds()
    #     elif q_num == 2:
    #         millis = 1000.0 * (event_client_time - self.latest_complete_lvl16).total_seconds()
    #     elif q_num == 3:
    #         millis = 1000.0 * (event_client_time - self.latest_answer_Q2).total_seconds()
    #     else:
    #         millis = None
    #     return int(millis) if millis is not None else None

    # ## Private function to do feature calculation at the end of a level.
    # #
    # #  @param lvl      The level whose features should be calculated
    # #  @return         (void)
    # def _calc_level_end(self, lvl:int) -> None:
    #     closenesses = self.move_closenesses_tx[lvl]['completeness']
    #     times = self.move_closenesses_tx[lvl]['t']
    #     ranges = self.move_closenesses_tx[lvl]['range']
    #     if times:
    #         try:
    #             X = [(times[i]-times[0]).seconds for i in range(len(times))]
    #         except Exception as err:
    #             Logger.Log(times[0])
    #             raise err
    #         y = closenesses
    #         if len(X) > 1:
    #             intercept, slope, r_sq = self._2D_linear_regression(X, y)
    #             r_sq = r_sq if not np.isnan(r_sq) else 0
    #         else:
    #             intercept, slope, r_sq = 0,0,0
    #         self._features.setValByIndex(feature_name='closenessIntercept', index=lvl, new_value=intercept)
    #         self._features.setValByIndex(feature_name='closenessSlope', index=lvl, new_value=slope)
    #         self._features.setValByIndex(feature_name='closenessR2', index=lvl, new_value=r_sq)

    #         if ranges[0] is None:
    #             del ranges[0]
    #             del times[0]

    #         X = [(times[i] - times[0]).seconds for i in range(len(times))]
    #         y = ranges
    #         if len(X) > 1:
    #             intercept, slope, r_sq = self._2D_linear_regression(X, y)
    #             r_sq = r_sq if not np.isnan(r_sq) else 0
    #         else:
    #             intercept, slope, r_sq = 0, 0, 0
    #         self._features.setValByIndex(feature_name='rangeIntercept', index=lvl, new_value=intercept)
    #         self._features.setValByIndex(feature_name='rangeSlope', index=lvl, new_value=slope)
    #         self._features.setValByIndex(feature_name='rangeR2', index=lvl, new_value=r_sq)


    # ## Private function to do feature calculation at the end of a level.
    # #
    # #  @param lvl      The level whose features should be calculated
    # #  @return         (void)
    # def _2D_linear_regression(self, x_vals:List[float], y_vals:List[float]):
    #     X, y = np.array(x_vals).reshape((-1,1)), np.array(y_vals)
    #     linreg = LinearRegression()
    #     linreg.fit(X, y)
    #     intercept, slope, r_sq = linreg.intercept_, linreg.coef_[0], linreg.score(X, y)
    #     return (intercept, slope, r_sq)



## import standard libraries
import abc
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Extractor import Extractor

## @class Model
#  Abstract base class for session-level Wave features.
#  Models only have one public function, called Eval.
#  The Eval function takes a list of row data, computes some statistic, and returns a list of results.
#  If the model works on features from session data, it should calculate one result for each row (each row being a session).
#  If the model works on a raw list of recent events, it should calculate a single result (each row being an event).
class Feature(Extractor):
    """
    Simple wrapper for the Extractor base class.
    
    Created for compatibility with existing feature extractors that use Feature as their base.
    as the same functions as Extractor class
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)




## import standard libraries
import abc
import logging
import typing
import logging
from collections import defaultdict
from datetime import timedelta
from typing import Any, Dict, List, Optional, Union
## import local files
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.schemas.games.GameSchema import GameSchema

LegacyFeatureType = Union[int,float,timedelta,Dict[int,Dict[str,Any]]]

## @class LegacyFeature
#  Abstract base class for game feature extractors.
#  Gives a few static functions to be used across all extractor classes,
#  and defines an interface that the SessionProcessor can use.
class LegacyFeature(Feature):

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _calculateAggregateFeatures(self) -> None:
        """Abstract declaration of a function to perform calculation of aggregate features
        from existing per-level/per-custom-count features.
        Will be called as a first step anytime GetFeatureValues is called.
        """
        return

    # *** BUILT-INS & PROPERTIES ***

    # Base constructor for LegacyFeature classes.
    def __init__(self, params:GeneratorParameters, game_schema:GameSchema, session_id:str):
        """Base constructor for LegacyFeature classes.
        The constructor sets an extractor's session id and range of levels,
        as well as initializing the features dictionary and list of played levels.

        :param session_id: The id of the session from which we will extract features.
        :type session_id: int
        :param game_schema: A dictionary that defines how the game data itself is structured
        :type game_schema: GameSchema
        """
        self._session_id  : str         = session_id
        self._game_schema : GameSchema  = game_schema
        self._levels      : List[int]   = []
        self._sequences   : List        = []
        self._features    : LegacyFeature.LegacySessionFeatures = LegacyFeature.LegacySessionFeatures(game_schema=game_schema)
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        self._calculateAggregateFeatures()
        def _format(obj):
            if obj == None:
                return ""
            elif type(obj) is timedelta:
                total_secs = obj.total_seconds()
                # h = total_secs // 3600
                # m = (total_secs % 3600) // 60
                # s = (total_secs % 3600) % 60 // 1  # just for reference
                # return f"{h:02.0f}:{m:02.0f}:{s:02.3f}"
                return str(total_secs)
            if obj is None:
                return ''
            else:
                return str(obj)
        column_vals = []
        # TODO: Should we do anything if the user accidentally adds a feature? For example I accidentally was adding 2
        # features that weren't in the schema (by misreferencing actual features), and they were appended to the end of
        # the feature list.
        for key in self._features.FeatureNames:
            _feature = self._features.getValByName(feature_name=key)
            if _feature is not None:
                key_type = type(_feature)
                if key_type is type({}) or key_type is type(defaultdict()):
                    # if it's a dictionary, expand.
                    column_vals.extend([_format(self._features.getValByIndex(key, num)) for num in _feature.keys()])
                else:
                    column_vals.append(_format(self._features.getValByName(key)))
        return column_vals

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***


    def GetFeatureNames(self) -> List[str]:
        columns = []
        for feat_name,feat_content in self._features.Features.items():
            if type(feat_content) is dict:
                # if it's a dictionary, expand.
                columns.extend([f"{feat_content[num]['prefix']}{num}_{feat_name}" for num in feat_content.keys()])
            else:
                columns.append(str(feat_name))
        return columns


    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    ## @class LegacySessionFeatures
    #  Private LegacyFeature class to track feature data.
    #  This class provides several functions to manage data, which should make
    #  the actual extractor code easier to read/write, since there is less need
    #  to understand the structure of feature data.
    class LegacySessionFeatures:
        def __init__(self, game_schema: GameSchema):
            self._features       : Dict[str, LegacyFeatureType] = {}
            self._perlevel_names : List[str] = list(game_schema._legacy_perlevel_feats.keys())

            _perlevels = game_schema._legacy_perlevel_feats
            _level_range = range(game_schema._min_level   if game_schema._min_level is not None else 0,
                                 game_schema._max_level+1 if game_schema._max_level is not None else 1)
            self._features.update({f:{lvl:{"val":None, "prefix":"lvl"} for lvl in _level_range } for f in _perlevels.keys()})
            # next, do something similar for other per-custom-count features.
            _percounts = game_schema.PerCountFeatures
            self._features.update({f:{num:{"val":None, "prefix":_percounts[f].Prefix} for num in range(0, int(_percounts[f].Count)) } for f in _percounts.keys()})
            # finally, add in aggregate-only features.
            self._features.update({f:0 for f in game_schema.AggregateFeatures.keys()})

        @property
        def Features(self):
            return self._features

        ## Getter function to retrieve a list of all features in the LegacySessionFeatures dictionary.
        #
        #  @return The keys in the LegacySessionFeatures dictionary.
        @property
        def FeatureNames(self):
            return self._features.keys()

        ## Function to initialize any previously uninitialized values of per-level features to 0, for given level.
        #  This means we can have "None" values for unreached levels, and 0's for features that
        #  simply never got incremented.
        #  @param level The level for which we should initialize values.
        def initLevel(self, level) -> None:
            for f_name in self._perlevel_names:
                feature = self._features[f_name]
                if type(feature) is dict and level in feature.keys():
                    if feature[level]["val"] == None:
                        feature[level]["val"] = 0
                else:
                    Logger.Log(f"Tried to initialize invalid level {level} for feature {f_name}", logging.ERROR)

        ## Function to get value of a per-count feature (including per-level)
        #  For a per-level feature, index is the level.
        #
        #  @param feature_name The name of the feature to retrieve
        #  @param index        The count index of the specific value, e.g. the level
        #  @return             The value stored for the given feature at given index.
        def getValByIndex(self, feature_name: str, index: int) -> Optional[Any]:
            if self._has_feature(feature_name):
                feature = self._features[feature_name]
                if type(feature) is dict and index in feature.keys():
                    return feature[index]["val"]
                else:
                    Logger.Log(f"Tried to get value on invalid index of {feature_name}: {index}", logging.ERROR)
            else:
                return None

        ## Function to get whole feature of a per-count feature (including per-level)
        #  For a per-level feature, index is the level.
        #
        #  @param feature_name The name of the feature to retrieve
        #  @param index        The count index of the desired value, e.g. the level
        #  @return             The feature stored for the given feature at given index.
        #                      This feature is a dictionary with a "val" and "prefix"
        def getFeatureByIndex(self, feature_name: str, index: int) -> Optional[Any]:
            if self._has_feature(feature_name):
                feature = self._features[feature_name]
                if type(feature) is dict and index in feature.keys():
                    return feature[index]
                else:
                    Logger.Log(f"Tried to get feature on invalid index of {feature_name}: {index}", logging.ERROR)
            else:
                return None

        ## Function to get all data on a given feature.
        #  Generally, this is intended for getting the value of an aggregate feature.
        #  However, it may also be used to get data across all levels/count for a
        #  per-count feature.
        #
        #  @param feature_name The name of the feature to retrieve
        #  @return             The value stored for the given feature.
        def getValByName(self, feature_name: str) -> Optional[Any]:
            if self._has_feature(feature_name):
                return self._features[feature_name]
            else:
                return None

        ## Function to set value of a per-count feature (including per-level)
        #  For a per-level feature, index is the level.
        #
        #  @param feature_name The name of the feature to set
        #  @param index        The count index of the desired value, e.g. the level
        #  @param new_value    The value to be stored for the given feature at given index.
        def setValByIndex(self, feature_name: str, index: int, new_value) -> None:
            if self._has_feature(feature_name):
                feature = self._features[feature_name]
                if type(feature) is dict and index in feature.keys():
                    feature[index]["val"] = new_value
                else:
                    Logger.Log(f"Tried to set value on invalid index of {feature_name}: {index}", logging.ERROR)

        ## Function to set value of a per-level feature
        #  Pure syntax sugar, calls setValByIndex using level as the index.
        #
        #  @param feature_name The name of the feature to set
        #  @param index        The count index of the desired value, e.g. the level
        #  @param new_value    The value to be stored for the given feature at given index.
        def setValByLevel(self, feature_name: str, level: int, new_value) -> None:
            self.setValByIndex(feature_name=feature_name, index=level, new_value=new_value)

        ## Function to set value of a full feature
        #  Intended for use with aggregate features. Not recommended for setting
        #  per-count features.
        #
        #  @param feature_name The name of the feature to retrieve
        #  @param new_value    The value to be stored for the given feature.
        def setValByName(self, feature_name: str, new_value) -> None:
            if self._has_feature(feature_name):
                self._features[feature_name] = new_value

        ## Function to increment the value of a per-count feature (including per-level)
        #  For a per-level feature, index is the level.
        #
        #  @param feature_name The name of the feature to increment
        #  @param index        The count index of the specific value, e.g. the level
        #  @param increment    The size of the increment (default = 1)
        def incValByIndex(self, feature_name: str, index: int, increment: Union[int, float] = 1) -> None:
            if self._has_feature(feature_name):
                feature = self._features[feature_name]
                if type(feature) is dict and index in feature.keys():
                    old_val = feature[index]["val"]
                    if old_val == 'null':
                        feature[index]["val"] = 0
                    elif (isinstance(old_val, int)) \
                      or (isinstance(old_val, float)):
                        feature[index]["val"] += increment
                    elif isinstance(old_val, timedelta):
                        feature[index]["val"] = old_val + timedelta(seconds=increment)
                else:
                    Logger.Log(f"Tried to increment value on invalid index of {feature_name}: {index}", logging.ERROR)

        ## Function to increment the value of a per-level feature
        #  Pure syntax sugar, calls incValByIndex using level as the index.
        #
        #  @param feature_name The name of the feature to increment
        #  @param index        The count index of the specific value, e.g. the level
        #  @param increment    The size of the increment (default = 1)
        def incValByLevel(self, feature_name: str, level: int, increment: Union[int, float] = 1) -> None:
            self.incValByIndex(feature_name=feature_name, index=level, increment=increment)

        ## Function to increment value of an aggregate feature
        #
        #  @param feature_name The name of the feature to increment
        #  @param increment    The size of the increment (default = 1)
        def incAggregateVal(self, feature_name: str, increment: Union[int, float] = 1) -> None:
            if self._has_feature(feature_name):
                old_val = self._features[feature_name]
                if (isinstance(old_val, int)) \
                or (isinstance(old_val, float)):
                    self._features[feature_name] = old_val + increment
                elif isinstance(old_val, timedelta):
                    self._features[feature_name] = old_val + timedelta(seconds=increment)
                else:
                    Logger.Log(f"In LegacyFeature, tried to increment {feature_name} of non-numeric type {type(self._features[feature_name])} by {increment} of type {type(increment)}", logging.WARN)
            else:
                Logger.Log("Attempted to increment a feature that doesn't exist!", logging.WARN)

        def _has_feature(self, feature_name) -> bool:
            try:
                _ = self._features[feature_name]
            except KeyError:
                Logger.Log(f'Feature {feature_name} does not exist.', logging.ERROR)
                return False
            return True



# import libraries
from typing import Any, Final, List, Optional
from ogd.core.generators.Generator import GeneratorParameters
# import local files
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class UsedContinue(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    START_SIGN : Final[str] = "tunic.historicalsociety.closet.gramps.intro_0_cs_0"

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._continue : int = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["FirstInteraction", "UsedSaveCode"] 

    def _updateFromEvent(self, event:Event) -> None:
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        if feature.FeatureType == "UsedSaveCode":
            self._save_code = feature.FeatureValues[0]
        elif feature.FeatureType == "FirstInteraction":
            self._first_inc = feature.FeatureValues[0]
        return

    def _getFeatureValues(self) -> List[Any]:
        if not self._save_code and self._first_inc != UsedContinue.START_SIGN:
            self._continue = 1
        return [self._continue]



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
# count of object_selected 
# between script_audio_started and script_audio_complete
# return the # of object_selected during audio playing
class ObjectSelectionsDuringVoiceover (SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._select_count = 0
        self._audio_started = False

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["script_audio_started","script_audio_complete","object_selected"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "script_audio_complete" or event.EventName == "scene_changed":
            self._audio_started = False
            return
        if not self._audio_started:
            if event.EventName == "script_audio_started":
                self._audio_started = True
            return
        if event.EventName == "object_selected":
            self._select_count += 1
        return


    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._select_count]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "2"


## import standard libraries
import logging
from typing import List, Set

# import local files
from ogd.core.interfaces.outerfaces.DataOuterface import DataOuterface
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

class DebugOuterface(DataOuterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, export_modes:Set[ExportMode]):
        super().__init__(game_id=game_id, export_modes=export_modes, config=config)
        # self.Open()

    def __del__(self):
        self.Close()

    # *** IMPLEMENT ABSTRACTS ***

    def _open(self) -> bool:
        self._display(f"Using a debug outerface to view OGD output for {self._game_id}.")
        return True

    def _close(self) -> bool:
        self._display(f"No longer using a debug outerface to view OGD output for {self._game_id}.")
        return True

    def _destination(self, mode:ExportMode) -> str:
        return "Logger.Log"

    def _removeExportMode(self, mode:ExportMode):
        match mode:
            case ExportMode.EVENTS:
                self._display("No longer outputting raw event data to debug stream.")
            case ExportMode.DETECTORS:
                self._display("No longer outputting processed event data to debug stream.")
            case ExportMode.SESSION:
                self._display("No longer outputting session data to debug stream.")
            case ExportMode.PLAYER:
                self._display("No longer outputting player data to debug stream.")
            case ExportMode.POPULATION:
                self._display("No longer outputting population data to debug stream.")

    def _writeRawEventsHeader(self, header:List[str]) -> None:
        self._display("Raw events header:")
        self._display(header)

    def _writeProcessedEventsHeader(self, header:List[str]) -> None:
        self._display("Processed events header:")
        self._display(header)

    def _writeSessionHeader(self, header:List[str]) -> None:
        self._display("Sessions header:")
        self._display(header)

    def _writePlayerHeader(self, header:List[str]) -> None:
        self._display("Player header:")
        self._display(header)

    def _writePopulationHeader(self, header:List[str]) -> None:
        self._display("Population header:")
        self._display(header)

    def _writeRawEventLines(self, events:List[ExportRow]) -> None:
        self._display("Raw event data:")
        _lengths = [len(elem) for elem in events]
        self._display(f"{len(events)} raw events, average length {sum(_lengths) / len(_lengths) if len(_lengths) > 0 else 'N/A'}")

    def _writeProcessedEventLines(self, events:List[ExportRow]) -> None:
        self._display("Processed event data:")
        _lengths = [len(elem) for elem in events]
        self._display(f"{len(events)} processed events, average length {sum(_lengths) / len(_lengths) if len(_lengths) > 0 else 'N/A'}")

    def _writeSessionLines(self, sessions:List[ExportRow]) -> None:
        self._display("Session data:")
        _lengths = [len(elem) for elem in sessions]
        self._display(f"{len(sessions)} events, average length {sum(_lengths) / len(_lengths) if len(_lengths) > 0 else 'N/A'}")

    def _writePlayerLines(self, players:List[ExportRow]) -> None:
        self._display("Player data:")
        _lengths = [len(elem) for elem in players]
        self._display(f"{len(players)} events, average length {sum(_lengths) / len(_lengths) if len(_lengths) > 0 else 'N/A'}")

    def _writePopulationLines(self, populations:List[ExportRow]) -> None:
        self._display("Population data:")
        _lengths = [len(elem) for elem in populations]
        self._display(f"{len(populations)} events, average length {sum(_lengths) / len(_lengths) if len(_lengths) > 0 else 'N/A'}")

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _display(self, msg):
        Logger.Log(f"DebugOuterface: {msg}", logging.DEBUG)



## import standard libraries
import json
import logging
from datetime import datetime
from typing import Any, Callable, List, Type, Optional, Set
## import local files
from ogd.core.generators.registries.DetectorRegistry import DetectorRegistry
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.processors.DetectorProcessor import DetectorProcessor
from ogd.core.processors.EventProcessor import EventProcessor
from ogd.core.models.Event import Event, EventSource
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.utils import utils
from ogd.core.utils.Logger import Logger
from ogd.core.utils.utils import ExportRow

## @class EventProcessor
#  Class to manage data for a csv events file.
class EventManager:
    def __init__(self, game_schema: GameSchema, trigger_callback:Callable[[Event], None],
                 LoaderClass:Optional[Type[GeneratorLoader]], feature_overrides:Optional[List[str]]=None):
        """Constructor for EventManager.
        Just creates empty list of lines and generates list of column names.
        """
        # define instance vars
        self._columns     : List[str]      = Event.ColumnNames()
        self._raw_events  : EventProcessor = EventProcessor(game_schema=game_schema)
        self._all_events  : EventProcessor = EventProcessor(game_schema=game_schema)
        self._detector_processor : Optional[DetectorProcessor] = None
        if LoaderClass is not None:
            self._detector_processor = DetectorProcessor(game_schema=game_schema,           LoaderClass=LoaderClass,
                                                         trigger_callback=trigger_callback, feature_overrides=feature_overrides)

    def ProcessEvent(self, event:Event, separator:str = "\t") -> None:
        # event.EventData = json.dumps(event.EventData)
        # TODO: double-check if the remote_addr is there to be dropped/ignored.
        self._all_events.ProcessEvent(event=event)
        if event.EventSource == EventSource.GAME:
            self._raw_events.ProcessEvent(event=event)
        if self._detector_processor is not None:
            self._detector_processor.ProcessEvent(event=event)

    def GetColumnNames(self) -> List[str]:
        return self._columns

    def GetRawLines(self, slice_num:int, slice_count:int) -> List[ExportRow]:
        start   : datetime = datetime.now()
        ret_val : List[Any] = self._raw_events.Lines
        time_delta = datetime.now() - start
        Logger.Log(f"Time to retrieve raw Event lines for slice [{slice_num}/{slice_count}]: {time_delta} to get {len(ret_val)} lines", logging.INFO, depth=2)
        return ret_val

    def GetAllLines(self, slice_num:int, slice_count:int) -> List[ExportRow]:
        start   : datetime = datetime.now()
        ret_val : List[Any] = self._all_events.Lines
        time_delta = datetime.now() - start
        Logger.Log(f"Time to retrieve all Event lines for slice [{slice_num}/{slice_count}]: {time_delta} to get {len(ret_val)} lines", logging.INFO, depth=2)
        return ret_val

    ## Function to empty the list of lines stored by the EventProcessor.
    #  This is helpful if we're processing a lot of data and want to avoid
    #  Eating too much memory.
    def ClearLines(self):
        Logger.Log(f"Clearing {len(self._raw_events.Lines)} raw event and {len(self._all_events.Lines)} processed event entries from EventManager.", logging.DEBUG, depth=2)
        self._raw_events.ClearLines()
        self._all_events.ClearLines()



'''
Example commands to be run in the commandline from the project directory:
py.test tests -s
pytest  tests/test_lakeland_models.py -s -k PopulationModel
pytest tests/test_lakeland_models.py::test_sequence_model[20070509155287116-PopulationModel] -vsl
pytest  tests/test_lakeland_models.py::test_sequence_model[20070513431426580-PopulationModel] --log-level=DEBUG
    --log-file=tests/logs/pop_negative_20070513431426580-PopulationModel.log -vl
    (logging is only helpful if there are logging.debug("") etc. statements in the model/test)

-v: verbose
-s: show print statements
-l: show local variables on error
-x: stop at first error
'''

import unittest
import logging
import pytest
import json
from tests import helpers
from realtime.ModelManager import ModelManager
from models.Model import ModelInputType


zip_path = lambda kind: f'tests/test_data/LAKELAND_20200501_to_20200530_5c141b6_{kind}.zip'
v18_zip_paths = lambda kind: f'tests/test_data/sample_v18_{kind}.zip'
kind_to_sep = {
    "dump": '\t',
    "proc": ',',
    "raw": '\t'
}

def get_df(kind, kind_to_path_func):
    df, _ = helpers.getLogDFbyPaths([kind_to_path_func(kind)],
                            index_cols=False,
                            sep=kind_to_sep[kind])
    if kind=="dump":
        df["event_data_complex"] = df["event_data_complex"].apply(lambda s: json.loads(s))
    return df

dfs = helpers.LazyDict(get_df, kind_to_path_func=zip_path)
v18_dfs = helpers.LazyDict(get_df, kind_to_path_func=v18_zip_paths)

proc_sessions = v18_dfs['proc'].to_dict('records')
proc_session_names = [f'{x["sessID"]}.{x["num_play"]}' for x in proc_sessions]
v18_dumps = [groupdf.to_dict('records') for _, groupdf in v18_dfs['dump'].groupby('sess_id')]
v18_dump_ids = [str(d[0]["sess_id"]) for d in v18_dumps]

model_mgr = ModelManager(game_name="LAKELAND")
model_names = model_mgr.ListModels()
all_models = [model_mgr.LoadModel(model_name) for model_name in model_names]
# print(*list(zip(model_mgr.ListModels(), all_models)), sep='\n')
feature_models, feature_model_names = zip(*[(model,model_name) for model,model_name in zip(all_models, model_names)
                  if model and model.GetInputType() == ModelInputType.FEATURE])
sequence_models, sequence_model_names = zip(*[(model,model_name) for model,model_name in zip(all_models, model_names)
                   if model and model.GetInputType() == ModelInputType.SEQUENCE])


@pytest.mark.parametrize("model", feature_models, ids=feature_model_names)
@pytest.mark.parametrize("proc_sessions", proc_sessions, ids=proc_session_names)
def test_feature_model(model,proc_sessions):
    if type(proc_sessions) is not list:
        proc_sessions = [proc_sessions]
    print(model.Eval(proc_sessions), end=', ')


@pytest.mark.parametrize("model", sequence_models, ids=sequence_model_names)
@pytest.mark.parametrize("session_dump", v18_dumps, ids=v18_dump_ids)
def test_sequence_model(model, session_dump):
    try:
        print(model.Eval(session_dump), end=', ')
    except AssertionError:
        logging.exception("Caught exception")
        raise


if __name__ == '__main__':
    pass



# import libraries
from os import truncate
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models import Event
from typing import Any, Final, List, Optional
# import locals
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class WorstPlayerAttribute(PerCountFeature):
    def __init__(self, params:GeneratorParameters):
        PerCountFeature.__init__(self, params=params)
        self._attr_count = 0
        self._ATTRIBUTE_ENUM : Final[List[str]] = ["endurance", "resourceful", "tech","social","trust","research"]
        self._attr_name = self._ATTRIBUTE_ENUM[self.CountIndex]


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event:Event):
        
        return False
        #return int(event.GameState['att']) == self.CountIndex



    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [""]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["WorstAttribute"]

    def _updateFromEvent(self, event:Event) -> None:
        #self._story_alignment = event.EventData["story_alignment"]

        pass


    def _updateFromFeatureData(self, feature:FeatureData):
        #add logic to make sure that MODE is session, not player so we don't get duplicates
        if(feature._mode == ExtractionMode.SESSION):
            attribute_list = feature._vals[1]
            for attr in attribute_list:
                if(self._ATTRIBUTE_ENUM.index(attr) == self.CountIndex):
                    self._attr_count+=1
            
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._attr_name, self._attr_count]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Count"]
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION] # >>> delete any modes you don't want run for your Feature. <<<
    



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.extractors.SessionFeature import SessionFeature


class RockMultiplePickupCount(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._current_count : int = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["peck_rock"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return [] 

    def _updateFromEvent(self, event:Event) -> None:
        # if has_rock does not exit it will return false as well
        if event.game_state.get("has_rock", False):
            self._current_count += 1
        

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._current_count]